# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 Jochen Sprickerhof <weechat@jochen.sprickerhof.de>
# Copyright (C) 2009-2013 Sebastien Helleu <flashcode@flashtux.org>
# Copyright (C) 2010 xt <xt@bash.no>
# Copyright (C) 2010 Aleksey V. Zapparov <ixti@member.fsf.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# whatsapp protocol for WeeChat.
# (this script requires WeeChat 0.3.0 (or newer) and the yowsup library)
#
# For help, see /help whatsapp
# Happy chat, enjoy :)
#
# History:
# 2015-12-30, Jochen Sprickerhof <weechat@jochen.sprickerhof.de>:
#     version 0.1: Reworked for Whatsapp
# 2013-09-30, Nils Görs <freenode.nils_2>:
#     version 1.6: add support of /secure for passwords and jid
#                : fix stdout/stderr when no JID was set
# 2013-05-14, Billiam <billiamthesecond@gmail.com>:
#     version 1.5: fix unicode encoding error in /jabber buddies
# 2013-05-03, Sebastien Helleu <flashcode@flashtux.org>:
#     version 1.4: add tags in user messages: notify_xxx, no_highlight,
#                  nick_xxx, prefix_nick_xxx, log1
# 2012-05-12, Sebastian Rydberg <sr@rydbergtech.se>:
#     version 1.3: Added support for fetching names from roster
# 2012-04-11, Sebastien Helleu <flashcode@flashtux.org>:
#     version 1.2: fix deletion of server options
# 2012-03-09, Sebastien Helleu <flashcode@flashtux.org>:
#     version 1.1: fix reload of config file
# 2012-01-03, Sebastien Helleu <flashcode@flashtux.org>:
#     version 1.0: changes for future compatibility with Python 3.x
# 2011-12-15, Sebastien Helleu <flashcode@flashtux.org>:
#     version 0.9: fix utf-8 encoding problem on jid
# 2011-03-21, Isaac Raway <isaac.raway@gmail.com>:
#     version 0.8: search chat buffer before opening it
# 2011-02-13, Sebastien Helleu <flashcode@flashtux.org>:
#     version 0.7: use new help format for command arguments
# 2010-11-23, xt
#     version 0.6: change format of sent ping, to match RFC
# 2010-10-05, xt, <xt@bash.no>
#     version 0.5: no highlight for status/presence messages
# 2010-10-01, xt, <xt@bash.no>
#     version 0.4:
#     add kick and invite
# 2010-08-03, Aleksey V. Zapparov <ixti@member.fsf.org>:
#     version 0.3:
#     add /jabber priority [priority]
#     add /jabber status [message]
#     add /jabber presence [online|chat|away|xa|dnd]
# 2010-08-02, Aleksey V. Zapparov <ixti@member.fsf.org>:
#     version 0.2.1:
#     fix prexence is set for current resource instead of sending
#         special presences for all buddies
# 2010-08-02, Aleksey V. Zapparov <ixti@member.fsf.org>:
#     version 0.2:
#     add priority and away_priority of resource
# 2010-08-02, Sebastien Helleu <flashcode@flashtux.org>:
#     version 0.1: first official version
# 2010-08-01, ixti <ixti@member.fsf.org>:
#     fix bug with non-ascii resources
# 2010-06-09, iiijjjiii <iiijjjiii@gmail.com>:
#     add connect server and port options (required for google talk)
#     add private option permitting messages to be displayed in separate
#       chat buffers or in a single server buffer
#     add jid aliases
#     add keepalive ping
# 2010-03-17, xt <xt@bash.no>:
#     add autoreconnect option, autoreconnects on protocol error
# 2010-03-17, xt <xt@bash.no>:
#     add autoconnect option, add new command /jmsg with -server option
# 2009-02-22, Sebastien Helleu <flashcode@flashtux.org>:
#     first version (unofficial)
#

SCRIPT_NAME    = "whatsapp"
SCRIPT_AUTHOR  = "Jochen Sprickerhof <weechat@jochen.sprickerhof.de>"
SCRIPT_VERSION = "0.1"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC    = "Whatsapp protocol for WeeChat"
SCRIPT_COMMAND = SCRIPT_NAME

import re

import_ok = True

try:
    import weechat
except:
    print("This script must be run under WeeChat.")
    print("Get WeeChat now at: http://www.weechat.org/")
    import_ok = False

try:
    from yowsup.common import YowConstants
    from yowsup.layers import YowLayerEvent
    from yowsup.layers.auth import AuthError
    from yowsup.layers.interface import YowInterfaceLayer, ProtocolEntityCallback
    from yowsup.layers.network import YowNetworkLayer
    from yowsup.layers.protocol_contacts.protocolentities.iq_statuses_get import GetStatusesIqProtocolEntity
    from yowsup.layers.protocol_contacts.protocolentities.iq_statuses_result import ResultStatusesIqProtocolEntity
    from yowsup.layers.protocol_iq import YowIqProtocolLayer
    from yowsup.layers.protocol_iq.protocolentities.iq import IqProtocolEntity
    from yowsup.layers.protocol_iq.protocolentities.iq_ping import PingIqProtocolEntity
    from yowsup.layers.protocol_messages.protocolentities.message_text import TextMessageProtocolEntity
    from yowsup.layers.protocol_presence.protocolentities.presence_available import AvailablePresenceProtocolEntity
    from yowsup.layers.protocol_presence.protocolentities.presence_subscribe import SubscribePresenceProtocolEntity
    from yowsup.layers.protocol_presence.protocolentities.presence_unavailable import UnavailablePresenceProtocolEntity
    from yowsup.layers.protocol_presence.protocolentities.presence_unsubscribe import UnsubscribePresenceProtocolEntity
    from yowsup.layers.protocol_profiles.protocolentities import SetStatusIqProtocolEntity
    from yowsup.stacks import YowStackBuilder
except:
    print("Package python-yowsup (yowsup) must be installed to use whatsapp protocol.")
    print("Get yowsup with your package manager, or at this URL: https://github.com/tgalal/yowsup")
    import_ok = False

# ==============================[ global vars ]===============================

whatsapp_servers = []
whatsapp_server_options = {
    "jid"          : { "type"         : "string",
                       "desc"         : "whatsapp id (user@server.tld)",
                       "min"          : 0,
                       "max"          : 0,
                       "string_values": "",
                       "default"      : "",
                       "value"        : "",
                       "check_cb"     : "",
                       "change_cb"    : "",
                       "delete_cb"    : "",
                       },
    "password"     : { "type"         : "string",
                       "desc"         : "password for whatsapp id on server",
                       "min"          : 0,
                       "max"          : 0,
                       "string_values": "",
                       "default"      : "",
                       "value"        : "",
                       "check_cb"     : "",
                       "change_cb"    : "",
                       "delete_cb"    : "",
                       },
    "autoconnect"  : { "type"         : "boolean",
                       "desc"         : "automatically connect to server when script is starting",
                       "min"          : 0,
                       "max"          : 0,
                       "string_values": "",
                       "default"      : "off",
                       "value"        : "off",
                       "check_cb"     : "",
                       "change_cb"    : "",
                       "delete_cb"    : "",
                       },
    "autoreconnect": { "type"         : "boolean",
                       "desc"         : "automatically reconnect to server when disconnected",
                       "min"          : 0,
                       "max"          : 0,
                       "string_values": "",
                       "default"      : "on",
                       "value"        : "on",
                       "check_cb"     : "",
                       "change_cb"    : "",
                       "delete_cb"    : "",
                       },
    "private"      : { "type"         : "boolean",
                       "desc"         : "display messages in separate chat buffers instead of a single server buffer",
                       "min"          : 0,
                       "max"          : 0,
                       "string_values": "",
                       "default"      : "on",
                       "value"        : "on",
                       "check_cb"     : "",
                       "change_cb"    : "",
                       "delete_cb"    : "",
                       },
    "recipes"      : { "type"         : "boolean",
                       "desc"         : "Send recipes for messages",
                       "min"          : 0,
                       "max"          : 0,
                       "string_values": "",
                       "default"      : "on",
                       "value"        : "on",
                       "check_cb"     : "",
                       "change_cb"    : "",
                       "delete_cb"    : "",
                       },
    "read"         : { "type"         : "boolean",
                       "desc"         : "Send read notifications",
                       "min"          : 0,
                       "max"          : 0,
                       "string_values": "",
                       "default"      : "on",
                       "value"        : "on",
                       "check_cb"     : "",
                       "change_cb"    : "",
                       "delete_cb"    : "",
                       },
    "ping_interval": { "type"         : "integer",
                       "desc"         : "Number of seconds between server pings. 0 = disable",
                       "min"          : 0,
                       "max"          : 9999999,
                       "string_values": "",
                       "default"      : "50",
                       "value"        : "50",
                       "check_cb"     : "ping_interval_check_cb",
                       "change_cb"    : "",
                       "delete_cb"    : "",
                       },
    "ping_timeout" : { "type"         : "integer",
                       "desc"         : "Number of seconds to allow ping to respond before timing out",
                       "min"          : 0,
                       "max"          : 9999999,
                       "string_values": "",
                       "default"      : "10",
                       "value"        : "10",
                       "check_cb"     : "ping_timeout_check_cb",
                       "change_cb"    : "",
                       "delete_cb"    : "",
                       },
    }
whatsapp_config_file = None
whatsapp_config_section = {}
whatsapp_config_option = {}
whatsapp_jid_aliases = {}             # { 'alias1': 'jid1', 'alias2': 'jid2', ... }

# =================================[ config ]=================================

def whatsapp_config_init():
    """ Initialize config file: create sections and options in memory. """
    global whatsapp_config_file, whatsapp_config_section
    whatsapp_config_file = weechat.config_new("whatsapp", "whatsapp_config_reload_cb", "")
    if not whatsapp_config_file:
        return
    # look
    whatsapp_config_section["look"] = weechat.config_new_section(
        whatsapp_config_file, "look", 0, 0, "", "", "", "", "", "", "", "", "", "")
    if not whatsapp_config_section["look"]:
        weechat.config_free(whatsapp_config_file)
        return
    whatsapp_config_option["debug"] = weechat.config_new_option(
        whatsapp_config_file, whatsapp_config_section["look"],
        "debug", "boolean", "display debug messages", "", 0, 0,
        "off", "off", 0, "", "", "", "", "", "")
    # color
    whatsapp_config_section["color"] = weechat.config_new_section(
        whatsapp_config_file, "color", 0, 0, "", "", "", "", "", "", "", "", "", "")
    if not whatsapp_config_section["color"]:
        weechat.config_free(whatsapp_config_file)
        return
    whatsapp_config_option["message_join"] = weechat.config_new_option(
        whatsapp_config_file, whatsapp_config_section["color"],
        "message_join", "color", "color for text in join messages", "", 0, 0,
        "green", "green", 0, "", "", "", "", "", "")
    whatsapp_config_option["message_quit"] = weechat.config_new_option(
        whatsapp_config_file, whatsapp_config_section["color"],
        "message_quit", "color", "color for text in quit messages", "", 0, 0,
        "red", "red", 0, "", "", "", "", "", "")
    # server
    whatsapp_config_section["server"] = weechat.config_new_section(
        whatsapp_config_file, "server", 0, 0,
        "whatsapp_config_server_read_cb", "", "whatsapp_config_server_write_cb", "",
        "", "", "", "", "", "")
    if not whatsapp_config_section["server"]:
        weechat.config_free(whatsapp_config_file)
        return
    whatsapp_config_section["jid_aliases"] = weechat.config_new_section(
        whatsapp_config_file, "jid_aliases", 0, 0,
        "whatsapp_config_jid_aliases_read_cb", "",
        "whatsapp_config_jid_aliases_write_cb", "",
        "", "", "", "", "", "")
    if not whatsapp_config_section["jid_aliases"]:
        weechat.config_free(whatsapp_config_file)
        return

def whatsapp_config_reload_cb(data, config_file):
    """ Reload config file. """
    return weechat.config_reload(config_file)

def whatsapp_config_server_read_cb(data, config_file, section, option_name, value):
    """ Read server option in config file. """
    global whatsapp_servers
    rc = weechat.WEECHAT_CONFIG_OPTION_SET_ERROR
    items = option_name.split(".", 1)
    if len(items) == 2:
        server = whatsapp_search_server_by_name(items[0])
        if not server:
            server = Server(items[0])
            whatsapp_servers.append(server)
            stackbuilder = YowStackBuilder()
            # disable status ping as weechat seems to have a problem with threads
            stackbuilder.setProp(YowIqProtocolLayer.PROP_PING_INTERVAL, 0)
            stackbuilder.pushDefaultLayers(True).push(server).build()
        if server:
            rc = weechat.config_option_set(server.options[items[1]], value, 1)
    return rc

def whatsapp_config_server_write_cb(data, config_file, section_name):
    """ Write server section in config file. """
    global whatsapp_servers
    weechat.config_write_line(config_file, section_name, "")
    for server in whatsapp_servers:
        for name, option in sorted(server.options.items()):
            weechat.config_write_option(config_file, option)
    return weechat.WEECHAT_RC_OK

def whatsapp_config_jid_aliases_read_cb(data, config_file, section, option_name, value):
    """ Read jid_aliases option in config file. """
    global whatsapp_jid_aliases
    whatsapp_jid_aliases[option_name] = value
    option = weechat.config_new_option(
        config_file, section,
        option_name, "string", "jid alias", "", 0, 0,
        "", value, 0, "", "", "", "", "", "")
    if not option:
        return weechat.WEECHAT_CONFIG_OPTION_SET_ERROR
    return weechat.WEECHAT_CONFIG_OPTION_SET_OK_CHANGED

def whatsapp_config_jid_aliases_write_cb(data, config_file, section_name):
    """ Write jid_aliases section in config file. """
    global whatsapp_jid_aliases
    weechat.config_write_line(config_file, section_name, "")
    for alias, jid in sorted(whatsapp_jid_aliases.items()):
        weechat.config_write_line(config_file, alias, jid)
    return weechat.WEECHAT_RC_OK

def whatsapp_config_read():
    """ Read whatsapp config file (whatsapp.conf). """
    global whatsapp_config_file
    return weechat.config_read(whatsapp_config_file)

def whatsapp_config_write():
    """ Write whatsapp config file (whatsapp.conf). """
    global whatsapp_config_file
    return weechat.config_write(whatsapp_config_file)

def whatsapp_debug_enabled():
    """ Return True if debug is enabled. """
    global whatsapp_config_options
    if weechat.config_boolean(whatsapp_config_option["debug"]):
        return True
    return False

def whatsapp_config_color(color):
    """ Return color code for a whatsapp color option. """
    global whatsapp_config_option
    if color in whatsapp_config_option:
        return weechat.color(weechat.config_color(whatsapp_config_option[color]))
    return ""

def ping_timeout_check_cb(server_name, option, value):
    global whatsapp_config_file, whatsapp_config_section
    ping_interval_option = weechat.config_search_option(
        whatsapp_config_file,
        whatsapp_config_section["server"],
        "%s.ping_interval" % (server_name)
        )
    ping_interval = weechat.config_integer(ping_interval_option)
    if int(ping_interval) and int(value) >= int(ping_interval):
        weechat.prnt("", "\nwhatsapp: unable to update 'ping_timeout' for server %s" % (server_name))
        weechat.prnt("", "whatsapp: to prevent multiple concurrent pings, ping_interval must be greater than ping_timeout")
        return weechat.WEECHAT_CONFIG_OPTION_SET_ERROR
    return weechat.WEECHAT_CONFIG_OPTION_SET_OK_CHANGED

def ping_interval_check_cb(server_name, option, value):
    global whatsapp_config_file, whatsapp_config_section
    ping_timeout_option = weechat.config_search_option(
        whatsapp_config_file,
        whatsapp_config_section["server"],
        "%s.ping_timeout" % (server_name)
        )
    ping_timeout = weechat.config_integer(ping_timeout_option)
    if int(value) and int(ping_timeout) >= int(value):
        weechat.prnt("", "\nwhatsapp: unable to update 'ping_interval' for server %s" % (server_name))
        weechat.prnt("", "whatsapp: to prevent multiple concurrent pings, ping_interval must be greater than ping_timeout")
        return weechat.WEECHAT_CONFIG_OPTION_SET_ERROR
    return weechat.WEECHAT_CONFIG_OPTION_SET_OK_CHANGED

# ================================[ servers ]=================================

class Server(YowInterfaceLayer):
    """ Class to manage a server: buffer, connection, send/recv data. """

    def __init__(self, name, **kwargs):
        """ Init server """
        global whatsapp_config_file, whatsapp_config_section, whatsapp_server_options

        super(Server, self).__init__()
        self.connected = False

        self.name = name
        # create options (user can set them with /set)
        self.options = {}
        # if the value is provided, use it, otherwise use the default
        values = {}
        for option_name, props in whatsapp_server_options.items():
            values[option_name] = props["default"]
        values['name'] = name
        values.update(**kwargs)
        for option_name, props in whatsapp_server_options.items():
            self.options[option_name] = weechat.config_new_option(
                whatsapp_config_file, whatsapp_config_section["server"],
                self.name + "." + option_name, props["type"], props["desc"],
                props["string_values"], props["min"], props["max"],
                props["default"], values[option_name], 0,
                props["check_cb"], self.name, props["change_cb"], "",
                props["delete_cb"], "")
        # internal data
        self.jid = None
        self.sock = None
        self.hook_fd = None
        self.buffer = ""
        self.chats = []
        self.buddies = []
        self.buddy = None
        self.ping_timer = None              # weechat.hook_timer for sending pings
        self.ping_timeout_timer = None      # weechat.hook_timer for monitoring ping timeout

    def option_string(self, option_name):
        """ Return a server option, as string. """
        return weechat.config_string(self.options[option_name])

    def option_boolean(self, option_name):
        """ Return a server option, as boolean. """
        return weechat.config_boolean(self.options[option_name])

    def option_integer(self, option_name):
        """ Return a server option, as string. """
        return weechat.config_integer(self.options[option_name])

    @ProtocolEntityCallback("receipt")
    def onReceipt(self, entity):
        self.toLower(entity.ack())

    @ProtocolEntityCallback("notification")
    def onNotification(self, notification):
        notificationData = notification.__str__()
        if notificationData:
            weechat.prnt('', "Notification: %s" % notificationData)
        else:
            weechat.prnt('', "From :%s, Type: %s" % (notification.getFrom(), notification.getType()))
        if weechat.config_boolean(self.options['recipes']):
            self.toLower(notification.ack())

    def connect(self):
        """ Connect to whatsapp server. """
        if not self.buffer:
            bufname = "%s.server.%s" % (SCRIPT_NAME, self.name)
            self.buffer = weechat.buffer_search("python", bufname)
            if not self.buffer:
                self.buffer = weechat.buffer_new(bufname,
                                                 "whatsapp_buffer_input_cb", "",
                                                 "whatsapp_buffer_close_cb", "")
            if self.buffer:
                weechat.buffer_set(self.buffer, "short_name", self.name)
                weechat.buffer_set(self.buffer, "localvar_set_type", "server")
                weechat.buffer_set(self.buffer, "localvar_set_server", self.name)
                weechat.buffer_set(self.buffer, "nicklist", "1")
                weechat.buffer_set(self.buffer, "nicklist_display_groups", "1")
                weechat.buffer_set(self.buffer, "display", "auto")

        self.buddy = Buddy(jid=eval_expression(self.option_string("jid")), server=self)

        credentials = (eval_expression(self.option_string("jid")), eval_expression(self.option_string("password")))
        self.getStack().setCredentials(credentials)
        self.getStack().broadcastEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_CONNECT))

        # set blocking, so we don't send in the select loop as asyncore does it
        self.sock = self.getStack().getLayer(0).socket.setblocking(1)
        # push initial connect message through the socket (would have been done in the select loop otherwise
        self.getStack().getLayer(0).handle_write_event()

        self.sock = self.getStack().getLayer(0).socket.fileno()
        self.hook_fd = weechat.hook_fd(self.sock, 1, 0, 0, "whatsapp_fd_cb", "")

        weechat.buffer_set(self.buffer, "highlight_words", self.buddy.alias)
        weechat.buffer_set(self.buffer, "localvar_set_nick", self.buddy.alias)
        hook_away = weechat.hook_command_run("/away -all*", "whatsapp_away_command_run_cb", "")

        # Whatsapp doesn't send context, so we use aliases as contacts instead
        for jid in whatsapp_jid_aliases.values():
            if jid != self.buddy.jid:
                self.add_buddy(jid)

    def is_connected(self):
        """Return connect status"""
        if not self.connected:
            return False
        else:
            return True

    def add_chat(self, buddy):
        """Create a chat buffer for a buddy"""
        chat = Chat(self, buddy, switch_to_buffer=False)
        self.chats.append(chat)
        return chat

    def del_buddy(self, jid):
        """ Remove a buddy and/or deny authorization request """
        entity = UnsubscribePresenceProtocolEntity(jid)
        self.toLower(entity)

    def print_debug_server(self, message):
        """ Print debug message on server buffer. """
        if whatsapp_debug_enabled():
            weechat.prnt(self.buffer, "%swhatsapp: %s" % (weechat.prefix("network"), message))

    def print_debug_handler(self, handler_name, node):
        """ Print debug message for a handler on server buffer. """
        self.print_debug_server("%s_handler, xml message:\n%s"
                                % (handler_name, node))

    def print_error(self, message):
        """ Print error message on server buffer. """
        if whatsapp_debug_enabled():
            weechat.prnt(self.buffer, "%swhatsapp: %s" % (weechat.prefix("error"), message))

    @ProtocolEntityCallback("chatstate")
    def presence_handler(self, node):
        print(node)
        print(dir(node))
        self.print_debug_handler("presence", node)
        buddy = self.search_buddy_list(node.getFrom(), by='jid')
        if not buddy:
            buddy = self.add_buddy(jid=node.getFrom())
        action='update'
        node_type = node.getType()
        if node_type in ["error", "unavailable"]:
            action='remove'
        if action == 'update':
            away = node.getShow() in ["away", "xa"]
            status = ' '
            if node.getStatus():
                status = node.getStatus()
            buddy.set_status(status=status, away=away)
        self.update_nicklist(buddy=buddy, action=action)
        return

    @ProtocolEntityCallback("iq")
    def iq_handler(self, node):
        """ Receive iq message. """
        self.print_debug_handler("iq", node)
        #weechat.prnt(self.buffer, "whatsapp: iq handler")
        if isinstance(node, ResultStatusesIqProtocolEntity):
            for jid in node.statuses:
                buddy = self.search_buddy_list(jid, by='jid')
                if not buddy:
                    buddy = self.add_buddy(jid=jid)
                buddy.set_status(status=node.statuses[jid][0])

        elif isinstance(node, IqProtocolEntity):
            self.delete_ping_timeout_timer()    # Disable the timeout feature
            if not self.is_connected() and weechat.config_boolean(self.options['autoreconnect']):
                self.connect()

    def onEvent(self, layerEvent):
        weechat.prnt('', layerEvent.getName())
        if layerEvent.getName() == YowNetworkLayer.EVENT_STATE_CONNECTED:
            self.connected = True
            return True
        elif layerEvent.getName() == YowNetworkLayer.EVENT_STATE_DISCONNECTED:
            weechat.prnt('', "Disconnected: %s" % layerEvent.getArg("reason"))
            if not self.is_connected() and weechat.config_boolean(self.options['autoreconnect']):
                self.connect()
            return True

    @ProtocolEntityCallback("message")
    def message_handler(self, node):
        """ Receive message. """
        self.print_debug_handler("message", node)
        messageOut = ""
        if node.getType() == "text":
            messageOut = self.getTextMessageBody(node)
        elif node.getType() == "media":
            messageOut = self.getMediaMessageBody(node)
        else:
            messageOut = "Unknown message type %s " % node.getType()
            self.print_debug_handler("send", messageOut.toProtocolTreeNode())

        jid = node.getFrom() if not node.isGroupMessage() else "%s/%s" % (node.getParticipant(False), node.getFrom())
        body = messageOut

        if weechat.config_boolean(self.options['recipes']):
            self.toLower(node.ack(weechat.config_boolean(self.options['recipes'])))
            self.print_debug_handler("Sent delivered receipt", "Message %s" % node.getId())

        if not jid or not body:
            return
        buddy = self.search_buddy_list(jid, by='jid')
        if not buddy:
            buddy = self.add_buddy(jid=jid)
        # If a chat buffer exists for the buddy, receive the message with that
        # buffer even if private is off. The buffer may have been created with
        # /query.
        recv_object = self
        if not buddy.chat and weechat.config_boolean(self.options['private']):
            self.add_chat(buddy)
        if buddy.chat:
            recv_object = buddy.chat
        recv_object.recv_message(buddy, body)

    def getTextMessageBody(self, message):
        return message.getBody()

    def getMediaMessageBody(self, message):
        if message.getMediaType() in ("image", "audio", "video"):
            return self.getDownloadableMediaMessageBody(message)
        else:
            return "Media Type: %s" % message.getMediaType()

    def getDownloadableMediaMessageBody(self, message):
        return "Media Type:{media_type}, Size:{media_size}, URL:{media_url}".format(
            media_type=message.getMediaType(),
            media_size=message.getMediaSize(),
            media_url=message.getMediaUrl()
        )

    def recv(self):
        """ Receive something from whatsapp server. """
        try:
            self.getStack().getLayer(0).handle_read()
        except AuthError as e:
            weechat.prnt('', '%s: Error from server: %s' %(SCRIPT_NAME, e))
            self.disconnect()
            if weechat.config_boolean(self.options['autoreconnect']):
                autoreconnect_delay = 30
                weechat.command('', '/wait %s /%s connect %s' %
                                (autoreconnect_delay, SCRIPT_COMMAND, self.name))

    def recv_message(self, buddy, message):
        """ Receive a message from buddy. """
        weechat.prnt_date_tags(self.buffer, 0,
                               "notify_private,nick_%s,prefix_nick_%s,log1" %
                               (buddy.alias,
                                weechat.config_string(weechat.config_get("weechat.color.chat_nick_other"))),
                               "%s%s\t%s" % (weechat.color("chat_nick_other"),
                                             buddy.alias,
                                             message))

    def print_status(self, nickname, status):
        """ Print a status in server window and in chat. """
        weechat.prnt_date_tags(self.buffer, 0, "no_highlight", "%s%s has status %s" %
                               (weechat.prefix("action"),
                                nickname,
                                status))
        for chat in self.chats:
            if nickname in chat.buddy.alias:
                chat.print_status(status)
                break

    def send_message(self, buddy, message):
        """ Send a message to buddy.

        The buddy argument can be either a jid string,
        eg username@domain.tld/resource or a Buddy object instance.
        """
        recipient = buddy
        if isinstance(buddy, Buddy):
            recipient = buddy.jid
        if not self.is_connected():
            weechat.prnt(self.buffer, "%swhatsapp: unable to send message, connection is down"
                         % weechat.prefix("error"))
            return
        outgoingMessage = TextMessageProtocolEntity(message, to=self.stringify_jid(recipient))
        self.toLower(outgoingMessage)

    def send_message_from_input(self, input=''):
        """ Send a message from input text on server buffer. """
        # Input must be of format "name: message" where name is a jid, bare_jid
        # or alias. The colon can be replaced with a comma as well.
        # Split input into name and message.
        if not re.compile(r'.+[:,].+').match(input):
            weechat.prnt(self.buffer, "%swhatsapp: %s" % (weechat.prefix("network"),
                "Invalid send format. Use  jid: message"
                ))
            return
        name, message = re.split('[:,]', input, maxsplit=1)
        buddy = self.search_buddy_list(name, by='alias')
        if not buddy:
            weechat.prnt(self.buffer,
                    "%swhatsapp: Invalid jid: %s" % (weechat.prefix("network"),
                    name))
            return
        # Send activity indicates user is no longer away, set it so
        if self.buddy and self.buddy.away:
            self.set_away('')
        self.send_message(buddy=buddy, message=message)
        try:
            sender = self.buddy.alias
        except:
            sender = self.jid
        weechat.prnt_date_tags(self.buffer, 0,
                               "notify_none,no_highlight,nick_%s,prefix_nick_%s,log1" %
                               (sender,
                                weechat.config_string(weechat.config_get("weechat.color.chat_nick_self"))),
                               "%s%s\t%s" % (weechat.color("chat_nick_self"),
                                             sender,
                                             message.strip()))

    def set_away(self, message):
        """ Set/unset away on server.

        If a message is provided, status is set to 'away'.
        If no message, then status is set to 'online'.
        """
        if message:
            entity = UnavailablePresenceProtocolEntity()
            self.toLower(entity)
        else:
            entity = AvailablePresenceProtocolEntity()
            self.toLower(entity)
        self.set_presence(message)

    def set_presence(self, status=None):
        message = status if status else ''
        entity = SetStatusIqProtocolEntity(message)
        self.toLower(entity)

    def add_buddy(self, jid):
        """ Add a new buddy """
        full_jid = self.stringify_jid(jid)
        entity = SubscribePresenceProtocolEntity(full_jid)
        self.toLower(entity)
        entity = GetStatusesIqProtocolEntity([full_jid])
        self.toLower(entity)

        buddy = Buddy(jid=jid, server=self)
        self.buddies.append(buddy)
        self.update_nicklist(buddy=buddy, action='update')
        return buddy

    def display_buddies(self):
        """ Display buddies. """
        weechat.prnt(self.buffer, "")
        weechat.prnt(self.buffer, "Buddies:")

        len_max = { 'alias': 5, 'jid': 5 }
        lines = []
        for buddy in sorted(self.buddies, key=lambda x: x.jid.getStripped()):
            alias = ''
            if buddy.alias != buddy.jid:
                alias = buddy.alias
            buddy_jid_string = buddy.jid.getStripped()
            lines.append( {
                'jid': buddy_jid_string,
                'alias': alias,
                'status': buddy.away_string(),
                })
            if len(alias) > len_max['alias']:
                len_max['alias'] = len(alias)
            if len(buddy_jid_string) > len_max['jid']:
                len_max['jid'] = len(buddy_jid_string)
        prnt_format = "  %s%-" + str(len_max['jid']) + "s %-" + str(len_max['alias']) + "s %s"
        weechat.prnt(self.buffer, prnt_format % ('', 'JID', 'Alias', 'Status'))
        for line in lines:
            weechat.prnt(self.buffer, prnt_format % (weechat.color("chat_nick"),
                                                    line['jid'],
                                                    line['alias'],
                                                    line['status'],
                                                    ))

    def stringify_jid(self, jid):
        """ Serialise JID into string.

        Args:
            jid: xmpp.protocol.JID, JID instance to serialize

        Notes:
            Method is based on original JID.__str__ but with hack to allow
            non-ascii in resource names.
        """
        if '@' in jid:
            return jid
        elif "-" in jid:
            return "%s@g.us" % jid

        return "%s@s.whatsapp.net" % jid

    def search_buddy_list(self, name, by='jid'):
        """ Search for a buddy by name.

        Args:
            name: string, the buddy name to search, eg the jid or alias
            by: string, either 'alias' or 'jid', determines which Buddy
                property to match on, default 'jid'

        Notes:
            If the 'by' parameter is set to 'jid', the search matches on all
            Buddy object jid properties, followed by all bare_jid properties.
            Once a match is found it is returned.

            If the 'by' parameter is set to 'alias', the search matches on all
            Buddy object alias properties.

            Generally, set the 'by' parameter to 'jid' when the jid is provided
            from a server, for example from a received message. Set 'by' to
            'alias' when the jid is provided by the user.
        """
        if by == 'jid':
            for buddy in self.buddies:
                if self.stringify_jid(buddy.jid) == name:
                    return buddy
            for buddy in self.buddies:
                if buddy.jid == name:
                    return buddy
        else:
            for buddy in self.buddies:
                if buddy.alias == name:
                    return buddy
        return None

    def update_nicklist(self, buddy=None, action=None):
        """Update buddy in nicklist
            Args:
                buddy: Buddy object instance
                action: string, one of 'update' or 'remove'
        """
        if not buddy:
            return
        if not action in ['remove', 'update']:
            return
        ptr_nick_gui = weechat.nicklist_search_nick(self.buffer, "", buddy.alias)
        weechat.nicklist_remove_nick(self.buffer, ptr_nick_gui)
        msg = ''
        prefix = ''
        color = ''
        away = ''
        if action == 'update':
            nick_color = "bar_fg"
            if buddy.away:
                nick_color = "weechat.color.nicklist_away"
            weechat.nicklist_add_nick(self.buffer, "", buddy.alias,
                                      nick_color, "", "", 1)
            if not ptr_nick_gui:
                msg = 'joined'
                prefix = 'join'
                color = 'message_join'
                away = buddy.away_string()
        if action == 'remove':
            msg = 'quit'
            prefix = 'quit'
            color = 'message_quit'
        if msg:
            weechat.prnt(self.buffer, "%s%s%s%s has %s %s"
                         % (weechat.prefix(prefix),
                            weechat.color("chat_nick"),
                            buddy.alias,
                            whatsapp_config_color(color),
                            msg,
                            away))
        return

    def add_ping_timer(self):
        if self.ping_timer:
            self.delete_ping_timer()
        if not self.option_integer('ping_interval'):
            return
        self.ping_timer = weechat.hook_timer( self.option_integer('ping_interval') * 1000,
                0, 0, "whatsapp_ping_timer", self.name)
        return

    def delete_ping_timer(self):
        if self.ping_timer:
            weechat.unhook(self.ping_timer)
        self.ping_timer = None
        return

    def add_ping_timeout_timer(self):
        if self.ping_timeout_timer:
            self.delete_ping_timeout_timer()
        if not self.option_integer('ping_timeout'):
            return
        self.ping_timeout_timer = weechat.hook_timer(
                self.option_integer('ping_timeout') * 1000, 0, 1,
                "whatsapp_ping_timeout_timer", self.name)
        return

    def delete_ping_timeout_timer(self):
        if self.ping_timeout_timer:
            weechat.unhook(self.ping_timeout_timer)
        self.ping_timeout_timer = None
        return

    def ping(self):
        if not self.is_connected():
            if not self.connect():
                return
        iq = PingIqProtocolEntity(to = YowConstants.DOMAIN)
        self.toLower(iq)
        self.print_debug_handler("ping", iq)
        self.add_ping_timeout_timer()
        return

    def ping_time_out(self):
        self.delete_ping_timeout_timer()
        # A ping timeout indicates a server connection problem. Disconnect
        # completely.
        self.disconnect()
        return

    def disconnect(self):
        """ Disconnect from whatsapp server. """
        if self.hook_fd != None:
            weechat.unhook(self.hook_fd)
            self.hook_fd = None
        self.getStack().broadcastEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_DISCONNECT))
        self.jid = None
        self.sock = None
        self.buddy = None
        weechat.nicklist_remove_all(self.buffer)

    def close_buffer(self):
        """ Close server buffer. """
        if self.buffer != "":
            weechat.buffer_close(self.buffer)
            self.buffer = ""

    def delete(self, deleteOptions=False):
        """ Delete server. """
        for chat in self.chats:
            chat.delete()
        self.delete_ping_timer()
        self.delete_ping_timeout_timer()
        self.disconnect()
        self.close_buffer()
        if deleteOptions:
            for name, option in self.options.items():
                weechat.config_option_free(option)

def eval_expression(option_name):
    """ Return a evaluated expression """
    if int(version) >= 0x00040200:
        return weechat.string_eval_expression(option_name,{},{},{})
    else:
        return option_name

def whatsapp_search_server_by_name(name):
    """ Search a server by name. """
    global whatsapp_servers
    for server in whatsapp_servers:
        if server.name == name:
            return server
    return None

def whatsapp_search_context(buffer):
    """ Search a server / chat for a buffer. """
    global whatsapp_servers
    context = { "server": None, "chat": None }
    for server in whatsapp_servers:
        if server.buffer == buffer:
            context["server"] = server
            return context
        for chat in server.chats:
            if chat.buffer == buffer:
                context["server"] = server
                context["chat"] = chat
                return context
    return context

def whatsapp_search_context_by_name(server_name):
    """Search for buffer given name of server. """

    bufname = "%s.server.%s" % (SCRIPT_NAME, server_name)
    return whatsapp_search_context(weechat.buffer_search("python", bufname))


# =================================[ chats ]==================================

class Chat:
    """ Class to manage private chat with buddy or MUC. """

    def __init__(self, server, buddy, switch_to_buffer):
        """ Init chat """
        self.server = server
        self.buddy = buddy
        buddy.chat = self
        bufname = "%s.%s.%s" % (SCRIPT_NAME, server.name, self.buddy.alias)
        self.buffer = weechat.buffer_search("python", bufname)
        if not self.buffer:
            self.buffer = weechat.buffer_new(bufname,
                                             "whatsapp_buffer_input_cb", "",
                                             "whatsapp_buffer_close_cb", "")
        self.buffer_title = self.buddy.alias
        if self.buffer:
            weechat.buffer_set(self.buffer, "title", self.buffer_title)
            weechat.buffer_set(self.buffer, "short_name", self.buddy.alias)
            weechat.buffer_set(self.buffer, "localvar_set_type", "private")
            weechat.buffer_set(self.buffer, "localvar_set_server", server.name)
            weechat.buffer_set(self.buffer, "localvar_set_channel", self.buddy.alias)
            weechat.hook_signal_send("logger_backlog",
                                     weechat.WEECHAT_HOOK_SIGNAL_POINTER, self.buffer)
            if switch_to_buffer:
                weechat.buffer_set(self.buffer, "display", "auto")

    def recv_message(self, buddy, message):
        """ Receive a message from buddy. """
        if buddy.alias != self.buffer_title:
            self.buffer_title = buddy.alias
            weechat.buffer_set(self.buffer, "title", "%s" % self.buffer_title)
        weechat.prnt_date_tags(self.buffer, 0,
                               "notify_private,nick_%s,prefix_nick_%s,log1" %
                               (buddy.alias,
                                weechat.config_string(weechat.config_get("weechat.color.chat_nick_other"))),
                               "%s%s\t%s" % (weechat.color("chat_nick_other"),
                                             buddy.alias,
                                             message))

    def send_message(self, message):
        """ Send message to buddy. """
        if not self.server.is_connected():
            weechat.prnt(self.buffer, "%swhatsapp: unable to send message, connection is down"
                         % weechat.prefix("error"))
            return
        self.server.send_message(self.buddy, message)
        weechat.prnt_date_tags(self.buffer, 0,
                               "notify_none,no_highlight,nick_%s,prefix_nick_%s,log1" %
                               (self.server.buddy.alias,
                                weechat.config_string(weechat.config_get("weechat.color.chat_nick_self"))),
                               "%s%s\t%s" % (weechat.color("chat_nick_self"),
                                             self.server.buddy.alias,
                                             message))
    def print_status(self, status):
        """ Print a status message in chat. """
        weechat.prnt(self.buffer, "%s%s has status %s" %
                     (weechat.prefix("action"),
                      self.buddy.alias,
                      status))

    def close_buffer(self):
        """ Close chat buffer. """
        if self.buffer != "":
            weechat.buffer_close(self.buffer)
            self.buffer = ""

    def delete(self):
        """ Delete chat. """
        self.close_buffer()

# =================================[ buddies ]==================================

class Buddy:
    """ Class to manage buddies. """
    def __init__(self, jid=None, chat=None, server=None ):
        """ Init buddy

        Args:
            jid: xmpp.protocol.JID object instance or string
            chat: Chat object instance
            server: Server object instance

        The jid argument can be provided either as a xmpp.protocol.JID object
        instance or as a string, eg "username@domain.tld/resource". If a string
        is provided, it is converted and stored as a xmpp.protocol.JID object
        instance.
        """

        # The jid argument of xmpp.protocol.JID can be either a string or a
        # xmpp.protocol.JID object instance itself.
        self.jid = jid
        self.chat = chat
        self.server = server
        self.name = ''
        self.alias = ''
        self.away = True
        self.status = ''

        self.set_alias()
        return

    def away_string(self):
        """ Return a string with away and status, with color codes. """
        if not self:
            return ''
        if not self.away:
            return ''
        str_colon = ": "
        if not self.status:
            str_colon = ""
        status = self.status.replace('\n', ' ') if self.status else ''
        return "%s(%saway%s%s%s)" % (weechat.color("chat_delimiters"),
                                      weechat.color("chat"),
                                      str_colon,
                                      status,
                                      weechat.color("chat_delimiters"))

    def set_alias(self):
        """Set the buddy alias.

        If an alias is defined in whatsapp_jid_aliases, it is used. Otherwise the
        alias is set to self.jid or self.name if it exists.
        """
        self.alias = self.jid
        if not self.jid:
            self.alias = ''
        if self.name:
            self.alias = self.name
        global whatsapp_jid_aliases
        for alias, jid in whatsapp_jid_aliases.items():
            if jid == self.jid:
                self.alias = alias
                break
        return

    def set_name(self, name=''):
        self.name = name
        self.set_alias()
        return

    def set_status(self, away=True, status=''):
        """Set the buddy status.

        Two properties define the buddy status.
            away - boolean, indicates whether the buddy is away or not.
            status - string, a message indicating the away status, eg 'in a meeting'
                   Comparable to xmpp presence <status/> element.
        """
        if not away and not status:
            status = 'online'
        # If the status has changed print a message on the server buffer
        if self.away != away or self.status != status:
            self.server.print_status(self.alias, status)
        self.away = away
        self.status = status
        return

# ================================[ commands ]================================

def whatsapp_hook_commands_and_completions():
    """ Hook commands and completions. """
    weechat.hook_command(SCRIPT_COMMAND, "Manage whatsapp servers",
                         "list || add <name> <jid> <password>"
                         " || connect|disconnect|del [<server>] || alias [add|del <alias> <jid>]"
                         " || away [<message>] || buddies ||"
                         " || status [<message>]"
                         " || debug || set <server> <setting> [<value>]",
                         "      list: list servers and chats\n"
                         "       add: add a server\n"
                         "   connect: connect to server using password\n"
                         "disconnect: disconnect from server\n"
                         "       del: delete server\n"
                         "     alias: manage jid aliases\n"
                         "      away: set away with a message (if no message, away is unset)\n"
                         "    status: set status message\n"
                         "   buddies: display buddies on server\n"
                         "     debug: toggle whatsapp debug on/off (for all servers)\n"
                         "\n"
                         "Without argument, this command lists servers and chats.\n"
                         "\n"
                         "Examples:\n"
                         "  Add a server:       /whatsapp add myserver user@server.tld password\n"
                         "  Add gtalk server:   /whatsapp add myserver user@gmail.com password talk.google.com:5223\n"
                         "  Connect to server:  /whatsapp connect myserver\n"
                         "  Disconnect:         /whatsapp disconnect myserver\n"
                         "  Delete server:      /whatsapp del myserver\n"
                         "\n"
                         "Aliases:\n"
                         "  List aliases:    /whatsapp alias \n"
                         "  Add an alias:    /whatsapp alias add alias_name jid\n"
                         "  Delete an alias: /whatsapp alias del alias_name\n"
                         "\n"
                         "Other whatsapp commands:\n"
                         "  Chat with a buddy (pv buffer): /query\n"
                         "  Add buddy to roster:           /winvite\n"
                         "  Remove buddy from roster:      /wkick\n"
                         "  Send message to buddy:         /wmsg",
                         "list %(whatsapp_servers)"
                         " || add %(whatsapp_servers)"
                         " || connect %(whatsapp_servers)"
                         " || disconnect %(whatsapp_servers)"
                         " || del %(whatsapp_servers)"
                         " || alias add|del %(whatsapp_jid_aliases)"
                         " || away"
                         " || status"
                         " || buddies"
                         " || debug",
                         "whatsapp_cmd_whatsapp", "")
    weechat.hook_command("query", "Chat with a whatsapp buddy",
                         "<buddy>",
                         "buddy: buddy id",
                         "",
                         "whatsapp_cmd_query", "")
    weechat.hook_command("wmsg", "Send a message to a buddy",
                         "[-server <server>] <buddy> <text>",
                         "server: name of whatsapp server buddy is on\n"
                         " buddy: buddy id\n"
                         "  text: text to send",
                         "",
                         "whatsapp_cmd_wmsg", "")
    weechat.hook_command("winvite", "Add a buddy to your roster",
                         "<buddy>",
                         "buddy: buddy id",
                         "",
                         "whatsapp_cmd_winvite", "")
    weechat.hook_command("wkick", "Remove a buddy from your roster, or deny auth",
                         "<buddy>",
                         "buddy: buddy id",
                         "",
                         "whatsapp_cmd_wkick", "")
    weechat.hook_completion("whatsapp_servers", "list of whatsapp servers",
                            "whatsapp_completion_servers", "")
    weechat.hook_completion("whatsapp_jid_aliases", "list of whatsapp jid aliases",
                            "whatsapp_completion_jid_aliases", "")

def whatsapp_list_servers_chats(name):
    """ List servers and chats. """
    global whatsapp_servers
    weechat.prnt("", "")
    if len(whatsapp_servers) > 0:
        weechat.prnt("", "whatsapp servers:")
        for server in whatsapp_servers:
            if name == "" or server.name.find(name) >= 0:
                conn_server = ''
                connected = ""
                if server.sock >= 0:
                    connected = "(connected)"

                weechat.prnt("", "  %s - %s %s %s" % (server.name,
                    eval_expression(server.option_string("jid")), conn_server, connected))
                for chat in server.chats:
                    weechat.prnt("", "    chat with %s" % (chat.buddy))
    else:
        weechat.prnt("", "whatsapp: no server defined")

def whatsapp_cmd_whatsapp(data, buffer, args):
    """ Command '/whatsapp'. """
    global whatsapp_servers, whatsapp_config_option
    if args == "" or args == "list":
        whatsapp_list_servers_chats("")
    else:
        argv = args.split(" ")
        argv1eol = ""
        pos = args.find(" ")
        if pos > 0:
            argv1eol = args[pos+1:]
        if argv[0] == "list":
            whatsapp_list_servers_chats(argv[1])
        elif argv[0] == "add":
            if len(argv) >= 4:
                server = whatsapp_search_server_by_name(argv[1])
                if server:
                    weechat.prnt("", "whatsapp: server '%s' already exists" % argv[1])
                else:
                    kwargs = {'jid': argv[2], 'password': argv[3]}
                    server = Server(argv[1], **kwargs)
                    whatsapp_servers.append(server)
                    stackbuilder = YowStackBuilder()
                    # disable status ping as weechat seems to have a problem with threads
                    stackbuilder.setProp(YowIqProtocolLayer.PROP_PING_INTERVAL, 0)
                    stackbuilder.pushDefaultLayers(True).push(server).build()
                    weechat.prnt("", "whatsapp: server '%s' created" % argv[1])
            else:
                weechat.prnt("", "whatsapp: unable to add server, missing arguments")
                weechat.prnt("", "whatsapp: usage: /whatsapp add name jid password")
        elif argv[0] == "alias":
            alias_command = AliasCommand(buffer, argv=argv[1:])
            alias_command.run()
        elif argv[0] == "connect":
            server = None
            if len(argv) >= 2:
                server = whatsapp_search_server_by_name(argv[1])
                if not server:
                    weechat.prnt("", "whatsapp: server '%s' not found" % argv[1])
            else:
                context = whatsapp_search_context(buffer)
                if context["server"]:
                    server = context["server"]
            if server:
                if weechat.config_boolean(server.options['autoreconnect']):
                    server.ping()               # This will connect and update ping status
                    server.add_ping_timer()
                else:
                    server.connect()
        elif argv[0] == "disconnect":
            server = None
            if len(argv) >= 2:
                server = whatsapp_search_server_by_name(argv[1])
                if not server:
                    weechat.prnt("", "whatsapp: server '%s' not found" % argv[1])
            else:
                context = whatsapp_search_context(buffer)
                if context["server"]:
                    server = context["server"]
            context = whatsapp_search_context(buffer)
            if server:
                server.delete_ping_timer()
                server.disconnect()
        elif argv[0] == "del":
            if len(argv) >= 2:
                server = whatsapp_search_server_by_name(argv[1])
                if server:
                    server.delete(deleteOptions=True)
                    whatsapp_servers.remove(server)
                    weechat.prnt("", "whatsapp: server '%s' deleted" % argv[1])
                else:
                    weechat.prnt("", "whatsapp: server '%s' not found" % argv[1])
        elif argv[0] == "send":
            if len(argv) >= 3:
                context = whatsapp_search_context(buffer)
                if context["server"]:
                    buddy = context['server'].search_buddy_list(argv[1], by='alias')
                    message = ' '.join(argv[2:])
                    context["server"].send_message(buddy, message)
        elif argv[0] == "read":
            whatsapp_config_read()
        elif argv[0] == "away":
            context = whatsapp_search_context(buffer)
            if context["server"]:
                context["server"].set_away(argv1eol)
        elif argv[0] == "status":
            context = whatsapp_search_context(buffer)
            if context["server"]:
                if len(argv) == 1:
                    weechat.prnt("", "whatsapp: status = %s" % context["server"].presence.getStatus())
                else:
                    context["server"].set_presence(status=argv1eol)
        elif argv[0] == "buddies":
            context = whatsapp_search_context(buffer)
            if context["server"]:
                context["server"].display_buddies()
        elif argv[0] == "debug":
            weechat.config_option_set(whatsapp_config_option["debug"], "toggle", 1)
            if whatsapp_debug_enabled():
                weechat.prnt("", "whatsapp: debug is now ON")
            else:
                weechat.prnt("", "whatsapp: debug is now off")
        else:
            weechat.prnt("", "whatsapp: unknown action")
    return weechat.WEECHAT_RC_OK

def whatsapp_cmd_query(data, buffer, args):
    """ Command '/query'. """
    if args:
        context = whatsapp_search_context(buffer)
        if context["server"]:
            buddy = context["server"].search_buddy_list(args, by='alias')
            if not buddy:
                buddy = context["server"].add_buddy(jid=args)
            if not buddy.chat:
                context["server"].add_chat(buddy)
            weechat.buffer_set(buddy.chat.buffer, "display", "auto")
    return weechat.WEECHAT_RC_OK

def whatsapp_cmd_wmsg(data, buffer, args):
    """ Command '/wmsg'. """
    if args:
        argv = args.split()
        if len(argv) < 2:
            return weechat.WEECHAT_RC_OK
        if argv[0] == '-server':
            context = whatsapp_search_context_by_name(argv[1])
            recipient = argv[2]
            message = " ".join(argv[3:])
        else:
            context = whatsapp_search_context(buffer)
            recipient = argv[0]
            message = " ".join(argv[1:])
        if context["server"]:
            buddy = context['server'].search_buddy_list(recipient, by='alias')
            context["server"].send_message(buddy, message)

    return weechat.WEECHAT_RC_OK

def whatsapp_cmd_winvite(data, buffer, args):
    """ Command '/winvite'. """
    if args:
        context = whatsapp_search_context(buffer)
        if context["server"]:
            context["server"].add_buddy(args)
    return weechat.WEECHAT_RC_OK

def whatsapp_cmd_wkick(data, buffer, args):
    """ Command '/wkick'. """
    if args:
        context = whatsapp_search_context(buffer)
        if context["server"]:
            context["server"].del_buddy(args)
    return weechat.WEECHAT_RC_OK

def whatsapp_away_command_run_cb(data, buffer, command):
    """ Callback called when /away -all command is run """
    global whatsapp_servers
    words = command.split(None, 2)
    if len(words) < 2:
        return
    message = ''
    if len(words) > 2:
        message = words[2]
    for server in whatsapp_servers:
        server.set_away(message)
    return weechat.WEECHAT_RC_OK

class AliasCommand(object):
    """Class representing a whatsapp alias command, ie /whatsapp alias ..."""

    def __init__(self, buffer, argv=None):
        """
        Args:
            bufffer: the weechat buffer the command was run in
            argv: list, the arguments provided with the command.
                  Example, if the command is "/whatsapp alias add abc abc@server.tld"
                  argv = ['add', 'abc', 'abc@server.tld']
        """
        self.buffer = buffer
        self.argv = []
        if argv:
            self.argv = argv
        self.action = ''
        self.jid = ''
        self.alias = ''
        self.parse()
        return

    def add(self):
        """Run a "/whatsapp alias add" command"""
        global whatsapp_jid_aliases
        if not self.alias or not self.jid:
            weechat.prnt("", "\nwhatsapp: unable to add alias, missing arguments")
            weechat.prnt("", "whatsapp: usage: /whatsapp alias add alias_name jid")
            return
        # Restrict the character set of aliases. The characters must be writable to
        # config file.
        invalid_re = re.compile(r'[^a-zA-Z0-9\[\]\\\^_\-{|}@\.]')
        if invalid_re.search(self.alias):
            weechat.prnt("", "\nwhatsapp: invalid alias: %s" % self.alias)
            weechat.prnt("", "whatsapp: use only characters: a-z A-Z 0-9 [ \ ] ^ _ - { | } @ .")
            return
        # Ensure alias and jid are reasonable length.
        max_len = 64
        if len(self.alias) > max_len:
            weechat.prnt("", "\nwhatsapp: invalid alias: %s" % self.alias)
            weechat.prnt("", "whatsapp: must be no more than %s characters long" % max_len)
            return
        if len(self.jid) > max_len:
            weechat.prnt("", "\nwhatsapp: invalid jid: %s" % self.jid)
            weechat.prnt("", "whatsapp: must be no more than %s characters long" % max_len)
            return
        jid = self.jid
        alias = self.alias
        if alias in whatsapp_jid_aliases.keys():
            weechat.prnt("", "\nwhatsapp: unable to add alias: %s" % (alias))
            weechat.prnt("", "whatsapp: alias already exists, delete first")
            return
        if jid in whatsapp_jid_aliases.values():
            weechat.prnt("", "\nwhatsapp: unable to add alias: %s" % (alias))
            for a, j in whatsapp_jid_aliases.items():
                if j == jid:
                    weechat.prnt("", "whatsapp: jid '%s' is already aliased as '%s', delete first" %
                        (j, a))
                    break
        whatsapp_jid_aliases[alias] = jid
        self.alias_reset(jid)
        return

    def alias_reset(self, jid):
        """Reset objects related to the jid modified by an an alias command

        Update any existing buddy objects, server nicklists, and chat objects
        that may be using the buddy with the provided jid.
        """
        global whatsapp_servers
        for server in whatsapp_servers:
            buddy = server.search_buddy_list(jid, by='jid')
            if not buddy:
                continue
            server.update_nicklist(buddy=buddy, action='remove')
            buddy.set_alias()
            server.update_nicklist(buddy=buddy, action='update')
            if buddy.chat:
                switch_to_buffer = False
                if buddy.chat.buffer == self.buffer:
                    switch_to_buffer = True
                buddy.chat.delete()
                new_chat = server.add_chat(buddy)
                if switch_to_buffer:
                    weechat.buffer_set(new_chat.buffer, "display", "auto")
        return

    def delete(self):
        """Run a "/whatsapp alias del" command"""
        global whatsapp_jid_aliases
        if not self.alias:
            weechat.prnt("", "\nwhatsapp: unable to delete alias, missing arguments")
            weechat.prnt("", "whatsapp: usage: /whatsapp alias del alias_name")
            return
        if not self.alias in whatsapp_jid_aliases:
            weechat.prnt("", "\nwhatsapp: unable to delete alias '%s', not found" % (self.alias))
            return
        jid = whatsapp_jid_aliases[self.alias]
        del whatsapp_jid_aliases[self.alias]
        self.alias_reset(jid)
        return

    def list(self):
        """Run a "/whatsapp alias" command to list aliases"""
        global whatsapp_jid_aliases
        weechat.prnt("", "")
        if len(whatsapp_jid_aliases) <= 0:
            weechat.prnt("", "whatsapp: no aliases defined")
            return
        weechat.prnt("", "whatsapp jid aliases:")
        len_alias = 5
        len_jid = 5
        for alias, jid in whatsapp_jid_aliases.items():
            if len_alias < len(alias):
                len_alias = len(alias)
            if len_jid < len(jid):
                len_jid = len(jid)
        prnt_format = "  %-" + str(len_alias) + "s %-" + str(len_jid) + "s"
        weechat.prnt("", prnt_format % ('Alias', 'JID'))
        for alias, jid in sorted(whatsapp_jid_aliases.items()):
            weechat.prnt("", prnt_format % (alias, jid))
        return

    def parse(self):
        """Parse the alias command into components"""
        if len(self.argv) <= 0:
            return
        self.action = self.argv[0]
        if len(self.argv) > 1:
            # Pad argv list to prevent IndexError exceptions
            while len(self.argv) < 3: self.argv.append('')
            self.alias = self.argv[1]
            self.jid = self.argv[2]
        return

    def run(self):
        """Execute the alias command."""
        if self.action == 'add':
            self.add()
        elif self.action == 'del':
            self.delete()
        self.list()
        return

def whatsapp_completion_servers(data, completion_item, buffer, completion):
    """ Completion with whatsapp server names. """
    global whatsapp_servers
    for server in whatsapp_servers:
        weechat.hook_completion_list_add(completion, server.name,
                                         0, weechat.WEECHAT_LIST_POS_SORT)
    return weechat.WEECHAT_RC_OK

def whatsapp_completion_jid_aliases(data, completion_item, buffer, completion):
    """ Completion with whatsapp alias names. """
    global whatsapp_jid_aliases
    for alias, jid in sorted(whatsapp_jid_aliases.items()):
        weechat.hook_completion_list_add(completion, alias,
                                         0, weechat.WEECHAT_LIST_POS_SORT)
    return weechat.WEECHAT_RC_OK

# ==================================[ fd ]====================================

def whatsapp_fd_cb(data, fd):
    """ Callback for reading socket. """
    global whatsapp_servers
    for server in whatsapp_servers:
        if server.sock == int(fd):
            server.recv()
    return weechat.WEECHAT_RC_OK

# ================================[ buffers ]=================================

def whatsapp_buffer_input_cb(data, buffer, input_data):
    """ Callback called for input data on a whatsapp buffer. """
    context = whatsapp_search_context(buffer)
    if context["server"] and context["chat"]:
        context["chat"].send_message(input_data)
    elif context["server"]:
        if input_data == "buddies" or "buddies".startswith(input_data):
            context["server"].display_buddies()
        else:
            context["server"].send_message_from_input(input=input_data)
    return weechat.WEECHAT_RC_OK

def whatsapp_buffer_close_cb(data, buffer):
    """ Callback called when a whatsapp buffer is closed. """
    context = whatsapp_search_context(buffer)
    if context["server"] and context["chat"]:
        if context["chat"].buddy:
            context["chat"].buddy.chat = None
        context["chat"].buffer = ""
        context["server"].chats.remove(context["chat"])
    elif context["server"]:
        context["server"].buffer = ""
    return weechat.WEECHAT_RC_OK

# ==================================[ timers ]==================================

def whatsapp_ping_timeout_timer(server_name, remaining_calls):
    server = whatsapp_search_server_by_name(server_name)
    if server:
        server.ping_time_out()
    return weechat.WEECHAT_RC_OK

def whatsapp_ping_timer(server_name, remaining_calls):
    server = whatsapp_search_server_by_name(server_name)
    if server:
        server.ping()
    return weechat.WEECHAT_RC_OK

# ==================================[ main ]==================================

if __name__ == "__main__" and import_ok:
    if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION,
                        SCRIPT_LICENSE, SCRIPT_DESC,
                        "whatsapp_unload_script", ""):

        version = weechat.info_get("version_number", "") or 0
        whatsapp_hook_commands_and_completions()
        whatsapp_config_init()
        whatsapp_config_read()
        for server in whatsapp_servers:
            if weechat.config_boolean(server.options['autoreconnect']):
                server.ping()               # This will connect and update ping status
                server.add_ping_timer()
            else:
                if weechat.config_boolean(server.options['autoconnect']):
                    server.connect()

# ==================================[ end ]===================================

def whatsapp_unload_script():
    """ Function called when script is unloaded. """
    global whatsapp_servers
    whatsapp_config_write()
    for server in whatsapp_servers:
        server.disconnect()
        server.delete()
    return weechat.WEECHAT_RC_OK



# -*- coding: utf-8 -*-
# Copyright (c) 2011-2014 by Jani Kesänen <jani.kesanen@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# A common buffer for URLs
#
# Collects received URLs from public and private messages into a single
# buffer. This buffer is especially handy if you spend lot's of time afk
# and you don't want to miss any of the cat pictures/videos that were pasted
# while you were doing something meaningful.
#
# This script has been originally developed for WeeChat version 0.3.5. May
# not work properly (or at all) on older versions.
#
# History:
# 2019-07-07, nils_2@freenode.#weechat
#   version 0.4: - fix bug when script unloads.
#                - add search for buffer name and display buffer name
# 2019-07-07, nils_2@freenode.#weechat
#   version 0.3: - make script compatible with Python 3.
# 2014-09-17, Jani Kesänen <jani.kesanen@gmail.com>
#   version 0.2: - added descriptions to settings.
# 2011-06-07, Jani Kesänen <jani.kesanen@gmail.com>
#   version 0.1: - initial release.
#

from __future__ import print_function

SCRIPT_NAME    = "urlbuf"
SCRIPT_AUTHOR  = "Jani Kesänen <jani.kesanen@gmail.com>"
SCRIPT_VERSION = "0.4"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC    = "A common buffer for received URLs."

import_ok = True

try:
    import weechat
except ImportError:
    print("This script must be run under WeeChat.")
    import_ok = False

import re

octet = r'(?:2(?:[0-4]\d|5[0-5])|1\d\d|\d{1,2})'
ipAddr = r'%s(?:\.%s){3}' % (octet, octet)
# Base domain regex off RFC 1034 and 1738
label = r'[0-9a-z][-0-9a-z]*[0-9a-z]?'
domain = r'%s(?:\.%s)*\.[a-z][-0-9a-z]*[a-z]?' % (label, label)
urlRe = re.compile(r'(\w+://(?:%s|%s)(?::\d+)?(?:/[^\])>\s]*)?)' % (domain, ipAddr), re.I)

urlbuf_buffer = None

urlbuf_settings = {
    "display_active_buffer" : ("on",  "display URLs from the active buffer"),
    "display_private"       : ("on",  "display URLs from private messages"),
    "display_buffer_number" : ("on",  "display the buffer's number or name (on/name/off)"),
    "display_nick"          : ("off", "display the nick of the user"),
    "skip_duplicates"       : ("on",  "skip the URL that is already in the urlbuf"),
    "skip_buffers"          : ("",    "a comma separated list of buffer numbers or buffer names to skip"),
}


def is_url_listed(buffer, url):
    """ Search for the URL from the buffer lines. """
    infolist = weechat.infolist_get("buffer_lines", buffer, "")

    found = False
    while weechat.infolist_next(infolist):
        message = weechat.infolist_string(infolist, "message").split(' ')[-1]
        if message == url:
            found = True
            break

    weechat.infolist_free(infolist)

    return found


def urlbuf_print_cb(data, buffer, date, tags, displayed, highlight, prefix, message):
    """ Called when a message is printed. """
    global urlbuf_buffer, urlbuf_tags

    # Exit immediately if the buffer does not exist
    if not urlbuf_buffer:
        return weechat.WEECHAT_RC_OK

    # Exit if the wanted tag is not in the message
    tagslist = tags.split(",")
    if not "notify_message" in tagslist:
        if weechat.config_get_plugin("display_private") == "on":
           if not "notify_private" in tagslist:
               return weechat.WEECHAT_RC_OK
        else:
           return weechat.WEECHAT_RC_OK

    # Exit if the message came from a buffer that is on the skip list
    buffer_number = str(weechat.buffer_get_integer(buffer, "number"))
    buffer_name = str(weechat.buffer_get_string(buffer, "name"))
    skips = set(weechat.config_get_plugin("skip_buffers").split(","))
    if buffer_number in skips:
        return weechat.WEECHAT_RC_OK
    if buffer_name in skips:
        return weechat.WEECHAT_RC_OK

    if weechat.config_get_plugin("display_active_buffer") == "off":
        if buffer_number == weechat.buffer_get_integer(weechat.current_buffer(), "number"):
            return weechat.WEECHAT_RC_OK

    # Process all URLs from the message
    for url in urlRe.findall(message):
        output = ""

        if weechat.config_get_plugin("skip_duplicates") == "on":
            if is_url_listed(urlbuf_buffer, url):
                continue

        if weechat.config_get_plugin("display_buffer_number") == "on":
            output += "%s%-2d " % (weechat.color("reset"), weechat.buffer_get_integer(buffer, "number"))
        elif weechat.config_get_plugin("display_buffer_number") == "name":
            output += "%s%s  " % (weechat.color("reset"), weechat.buffer_get_string(buffer, "name"))

        if weechat.config_get_plugin("display_nick") == "on":
            output += "%s " % (prefix)

        # Output the formatted URL into the buffer
        weechat.prnt(urlbuf_buffer, output + url)

    return weechat.WEECHAT_RC_OK


def urlbuf_input_cb(data, buffer, input_data):
    """ A Dummy callback for buffer input. """
    return weechat.WEECHAT_RC_OK


def urlbuf_close_cb(data, buffer):
    """ A callback for buffer closing. """
    global urlbuf_buffer

    urlbuf_buffer = None
    return weechat.WEECHAT_RC_OK

def urlbuf2_close_cb():
    global urlbuf_buffer
    urlbuf_buffer = None
    return weechat.WEECHAT_RC_OK

if __name__ == "__main__" and import_ok:
    if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION,
                        SCRIPT_LICENSE, SCRIPT_DESC, "urlbuf2_close_cb", ""):
        version = weechat.info_get('version_number', '') or 0

        # Set default settings
        for option, default_value in urlbuf_settings.items():
            if not weechat.config_is_set_plugin(option):
                weechat.config_set_plugin(option, default_value[0])
            if int(version) >= 0x00030500:
                weechat.config_set_desc_plugin(option, default_value[1])

        urlbuf_buffer = weechat.buffer_search("python", "urlbuf")

        if not urlbuf_buffer:
            # Create urlbuf. Sets notify to 0 as this buffer does not need to
            # be in hotlist.
            urlbuf_buffer = weechat.buffer_new("urlbuf", "urlbuf_input_cb", \
                                               "", "urlbuf_close_cb", "")
            weechat.buffer_set(urlbuf_buffer, "title", "URL buffer")
            weechat.buffer_set(urlbuf_buffer, "notify", "0")
            weechat.buffer_set(urlbuf_buffer, "nicklist", "0")

        # Hook all public and private messages (some may think this is too limiting)
        weechat.hook_print("", "notify_message", "", 1, "urlbuf_print_cb", "")
        weechat.hook_print("", "notify_private", "", 1, "urlbuf_print_cb", "")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2019 Sébastien Helleu <flashcode@flashtux.org>
# Copyright (C) 2009 drubin <drubin [@] smartcube [dot] co[dot]za>
# Copyright (C) 2010-2011 Trashlord <dornenreich666@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# Tetris game for WeeChat.
#
# History:
#
# 2019-09-29, Sébastien Helleu <flashcode@flashtux.org>:
#     version 1.0:
#       - convert script from Perl to Python
#       - use 256 colors if available
#       - use original Tetris colors for pieces
#       - add input action "q" to close weetris buffer
#       - display next piece (new option "display_next_piece")
#       - add key ctrl-down: move to bottom (new option "key_down_slow")
# 2011-02-28, Trashlord <dornenreich666@gmail.com>:
#     version 0.9: add playing time display
# 2010-10-08, Trashlord <dornenreich666@gmail.com>:
#     version 0.8: add best score and best level statistics
# 2009-12-17, Sébastien Helleu <flashcode@flashtux.org>:
#     version 0.7: add levels, fix bugs with pause
# 2009-12-16, drubin <drubin [@] smartcube [dot] co[dot]za>:
#     version 0.6: add key for pause, basic doc and auto jump to buffer
# 2009-06-21, Sébastien Helleu <flashcode@flashtux.org>:
#     version 0.5: fix bug with weetris buffer after /upgrade
# 2009-05-02, Sébastien Helleu <flashcode@flashtux.org>:
#     version 0.4: sync with last API changes, fix problem with key alt-n
# 2008-11-14, Sébastien Helleu <flashcode@flashtux.org>:
#     version 0.3: minor code cleanup
# 2008-11-12, Sébastien Helleu <flashcode@flashtux.org>:
#     version 0.2: hook timer only when weetris buffer is open
# 2008-11-05, Sébastien Helleu <flashcode@flashtux.org>:
#     version 0.1: first official version
# 2008-04-30, Sébastien Helleu <flashcode@flashtux.org>:
#     script creation

"""Tetris game for WeeChat, yeah!"""

from __future__ import print_function

SCRIPT_NAME = 'weetris'
SCRIPT_AUTHOR = 'Sébastien Helleu <flashcode@flashtux.org>'
SCRIPT_VERSION = '1.0'
SCRIPT_LICENSE = 'GPL3'
SCRIPT_DESC = 'Tetris game for WeeChat, yeah!'

import_ok = True

try:
    import weechat
except ImportError:
    print('This script must be run under WeeChat.')
    print('Get WeeChat now at: https://weechat.org/')
    import_ok = False

try:
    import random
    import time
except ImportError as message:
    print('Missing package(s) for %s: %s' % (SCRIPT_NAME, message))
    import_ok = False

BUFFER_TITLE = ('%s %s - enjoy!  |  '
                'Keys: arrows: move/rotate (ctrl-down: bottom), '
                'alt-N: new game, alt-P: pause  |  '
                'Input: q = quit' % (SCRIPT_NAME, SCRIPT_VERSION))

GAME_WIDTH = 10
GAME_HEIGHT = 20

MAX_LEVEL = 10

START_Y = 0

PIECES = (
    1024+512+64+32,     # O
    2048+1024+512+256,  # I
    2048+1024+512+64,   # T
    2048+1024+512+128,  # L
    2048+1024+512+32,   # J
    1024+512+128+64,    # S
    2048+1024+64+32,    # Z
)
PIECE_COLOR = {
    # < 256 colors
    False: [
        'yellow',     # O
        'lightcyan',  # I
        'magenta',    # T
        'brown',      # L
        'blue',       # J
        'green',      # S
        'red',        # Z

    ],
    # >= 256 colors
    True: [
        'yellow',     # O
        'lightcyan',  # I
        'magenta',    # T
        '172',        # L
        'blue',       # J
        'green',      # S
        'red',        # Z
    ],
}
PIECE_X_INC = (3, 2, 1, 0, 3, 2, 1, 0, 3, 2, 1, 0, 3, 2, 1, 0)
PIECE_Y_INC = (3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0)
PIECE_ROTATION = (
    4096, 256, 16, 1,
    8192, 512, 32, 2,
    16384, 1024, 64, 4,
    32768, 2048, 128, 8,
)

# -------------------------          -------------------------
# |     |     |     |     |          |     |     |     |     |
# |32768|16384| 8192| 4096|          |  8  | 128 | 2048|32768|
# |     |     |     |     |          |     |     |     |     |
# -------------------------          -------------------------
# |     |     |     |     |          |     |     |     |     |
# | 2048| 1024| 512 | 256 |          |  4  |  64 | 1024|16384|
# |     |     |     |     |  after   |     |     |     |     |
# -------------------------  rotate  -------------------------
# |     |     |     |     |   ===>   |     |     |     |     |
# | 128 |  64 |  32 |  16 |          |  2  |  32 | 512 | 8192|
# |     |     |     |     |          |     |     |     |     |
# -------------------------          -------------------------
# |     |     |     |     |          |     |     |     |     |
# |  8  |  4  |  2  |  1  |          |  1  |  16 | 256 | 4096|
# |     |     |     |     |          |     |     |     |     |
# -------------------------          -------------------------

weetris = {
    # will be set to True if 256 colors are supported by the terminal
    '256colors': False,
    'buffer': '',
    'timer': '',
    'level': 1,
    'matrix': [],
    'matrix_next': [],
    'playing': False,
    'paused': False,
    'lines': 0,
    'piece_x': 0,
    'piece_y': 0,
    'piece_number': -1,
    'next_piece_number': -1,
    'piece_form': 0,
    'best_level': 1,
    'best_lines': 0,
    'play_start_time': 0,
    'time_display_timer': '',
}

# script options
weetris_settings_default = {
    'display_next_piece': (
        'on',
        'display the next piece',
    ),
    'key_down_slow': (
        'on',
        'the key "down" moves the piece slowly: one position, and ctrl-down '
        'mores directly to the bottom; if disabled, the two keys are '
        'reversed',
    ),
}
weetris_settings = {}


def weetris_config_cb(data, option, value):
    """Called when a script option is changed."""
    pos = option.rfind('.')
    if pos > 0:
        name = option[pos+1:]
        if name in weetris_settings:
            weetris_settings[name] = value
    return weechat.WEECHAT_RC_OK


def buffer_input_cb(data, buf, input_data):
    """Input on weetris buffer."""
    if input_data == 'q':
        weechat.buffer_close(weetris['buffer'])
    return weechat.WEECHAT_RC_OK


def buffer_close_cb(data, buf):
    """Weetris buffer closed (oh no, why?)."""
    weetris['buffer'] = ''
    if weetris['timer']:
        weechat.unhook(weetris['timer'])
        weetris['timer'] = ''
    if weetris['time_display_timer']:
        weechat.unhook(weetris['time_display_timer'])
        weetris['play_start_time'] = 0
        weetris['time_display_timer'] = ''
    weechat.prnt('', 'Thank you for playing WeeTris!')
    return weechat.WEECHAT_RC_OK


def get_piece_block(value):
    """
    Return a string with a single block of a piece
    (spaces with background color).
    """
    if value < 0:
        block = weechat.color(',default')
    else:
        block = weechat.color(',' + PIECE_COLOR[weetris['256colors']][value])
    return block + '  '


def display_line(y):
    """Display a line of the matrix."""
    line = ' │'
    if weetris['paused']:
        if y == GAME_HEIGHT // 2:
            spaces_before = ((GAME_WIDTH * 2) - 6) // 2
            spaces_after = (GAME_WIDTH * 2) - 6 - spaces_before
            line += (' ' * spaces_before) + 'PAUSED' + (' ' * spaces_after)
        else:
            line += '  ' * GAME_WIDTH
    else:
        for x in range(GAME_WIDTH):
            line += get_piece_block(weetris['matrix'][y][x])
    line += weechat.color(',default') + '│'
    if weetris['playing'] and weetris_settings['display_next_piece'] == 'on':
        if y == 0:
            line += '    Next: '
        elif 1 <= y <= 4:
            line += '    '
            for x in range(4):
                line += get_piece_block(weetris['matrix_next'][y - 1][x])
            line += weechat.color(',default')
    weechat.prnt_y(weetris['buffer'], START_Y + y + 1, line)


def weetris_display_playing_time_cb(data, remaining_calls):
    """Callback of timer to display the playing time."""
    total_seconds = time.time() - weetris['play_start_time']
    minutes = int(total_seconds // 60)
    seconds = int(total_seconds % 60)
    total_seconds += 1
    weechat.prnt_y(weetris['buffer'], START_Y + GAME_HEIGHT + 6,
                   ' Playing time : %02d:%02d' % (minutes, seconds))
    return weechat.WEECHAT_RC_OK


def display_level_lines():
    """Display the current level and number of lines."""
    list_info = [
        'Level %-3d %6d line%s' % (weetris['level'],
                                   weetris['lines'],
                                   's' if weetris['lines'] > 1 else ''),
        '-' * (1 + (GAME_WIDTH * 2) + 1),
        'Highest level: %d' % weetris['best_level'],
        'Max lines    : %d' % weetris['best_lines'],
    ]
    for y, info in enumerate(list_info):
        weechat.prnt_y(weetris['buffer'], START_Y + GAME_HEIGHT + 2 + y,
                       ' ' + info)


def display_piece(display):
    """Display (or hide) the current piece."""
    value = weetris['piece_number'] if display else -1
    for i in range(16):
        if weetris['piece_form'] & (1 << i):
            x2 = weetris['piece_x'] + PIECE_X_INC[i]
            y2 = weetris['piece_y'] + PIECE_Y_INC[i]
            weetris['matrix'][y2][x2] = value


def display_all():
    """Display everything on the weetris buffer."""
    display_piece(True)
    weechat.prnt_y(weetris['buffer'], START_Y,
                   ' ┌' + ('──' * GAME_WIDTH) + '┐')
    for y in range(GAME_HEIGHT):
        display_line(y)
    weechat.prnt_y(weetris['buffer'], START_Y + GAME_HEIGHT + 1,
                   ' └' + ('──' * GAME_WIDTH) + '┘')
    display_piece(False)


def random_piece():
    """Return a random piece number."""
    return random.randint(0, len(PIECES) - 1)


def set_matrix_next():
    """Set the matrix for the next piece."""
    weetris['matrix_next'] = [[-1] * 4 for i in range(4)]
    number = weetris['next_piece_number']
    form = PIECES[number]
    for i in range(16):
        if form & (1 << i):
            weetris['matrix_next'][PIECE_Y_INC[i]][PIECE_X_INC[i]] = number


def set_new_form():
    """Choose a new random form."""
    if weetris['next_piece_number'] < 0:
        weetris['next_piece_number'] = random_piece()
    weetris['piece_number'] = weetris['next_piece_number']
    weetris['next_piece_number'] = random_piece()
    set_matrix_next()
    weetris['piece_form'] = PIECES[weetris['piece_number']]
    weetris['piece_x'] = (GAME_WIDTH // 2) - 2
    weetris['piece_y'] = 0


def init_timer():
    """Initialize timer."""
    if weetris['timer']:
        weechat.unhook(weetris['timer'])
    delay = max(100, 700 - ((weetris['level'] - 1) * 60))
    weetris['timer'] = weechat.hook_timer(delay, 0, 0, 'weetris_timer_cb', '')


def new_game():
    """New game."""
    weechat.prnt_y(weetris['buffer'], START_Y + GAME_HEIGHT + 2, '')
    weetris['matrix'] = [[-1] * GAME_WIDTH for i in range(GAME_HEIGHT)]
    weetris['next_piece_number'] = -1
    set_new_form()
    weetris['playing'] = True
    weetris['paused'] = False
    weetris['lines'] = 0
    weetris['level'] = 1
    weetris['play_start_time'] = time.time()
    weechat.prnt_y(weetris['buffer'], START_Y + GAME_HEIGHT + 6,
                   ' Playing time : 00:00')
    init_timer()
    weetris['time_display_timer'] = weechat.hook_timer(
        1000, 0, 0,
        'weetris_display_playing_time_cb', '',
    )
    display_all()
    display_level_lines()


def rotation(form):
    """Rotate a form."""
    new_form = 0
    for i in range(16):
        if form & (1 << i):
            new_form |= PIECE_ROTATION[i]
    return new_form


def is_possible(new_x, new_y, new_form):
    """Check if the "new_form" can be moved to position (new_x, new_y)."""
    for i in range(16):
        if not new_form & (1 << i):
            continue
        x = new_x + PIECE_X_INC[i]
        y = new_y + PIECE_Y_INC[i]
        if x < 0 or x >= GAME_WIDTH or y < 0 or y >= GAME_HEIGHT \
                or weetris['matrix'][y][x] >= 0:
            return 0
    return 1


def remove_completed_lines():
    """Remove completed lines."""
    y = GAME_HEIGHT - 1
    lines_removed = False
    while y >= 0:
        if -1 not in weetris['matrix'][y]:
            for i in range(y, -1, -1):
                if i == 0:
                    weetris['matrix'][i] = [-1] * GAME_WIDTH
                else:
                    weetris['matrix'][i] = weetris['matrix'][i - 1]
            # Removes the line and increases the number of lines made
            # in the game in "lines"
            weetris['lines'] += 1
            lines_removed = True
            if weetris['lines'] > weetris['best_lines']:
                set_best('max_lines', weetris['lines'])
                weetris['best_lines'] = weetris['lines']
        else:
            y -= 1
    if lines_removed:
        new_level = min(MAX_LEVEL, (weetris['lines'] // 10) + 1)
        if new_level != weetris['level']:
            # Next level
            weetris['level'] = new_level
            if weetris['level'] > weetris['best_level']:
                set_best('max_level', weetris['level'])
                weetris['best_level'] = weetris['level']
            init_timer()
        display_level_lines()


def end_of_piece():
    """End of a piece (it can not go down any more)."""
    display_piece(True)
    set_new_form()
    if is_possible(weetris['piece_x'], weetris['piece_y'],
                   weetris['piece_form']):
        remove_completed_lines()
    else:
        weetris['piece_form'] = 0
        weetris['playing'] = False
        weetris['paused'] = False
        if weetris['time_display_timer']:
            weechat.unhook(weetris['time_display_timer'])
            weetris['time_display_timer'] = ''
        weechat.prnt_y(weetris['buffer'], START_Y + GAME_HEIGHT + 2,
                       '>> End of game, score: %d lines, level %d '
                       '(alt-N to restart) <<' % (weetris['lines'],
                                                  weetris['level']))


def weetris_init():
    """Initialize weetris."""
    keys = {
        'meta2-A': 'up',
        'meta2-B': 'down',
        'meta-Ob': 'bottom',
        'meta-OB': 'bottom',
        'meta2-1;5B': 'bottom',
        'meta2-D': 'left',
        'meta2-C': 'right',
        'meta-n': 'new_game',
        'meta-p': 'pause',
    }
    weetris['buffer'] = weechat.buffer_search('python', 'weetris')
    if not weetris['buffer']:
        weetris['buffer'] = weechat.buffer_new('weetris',
                                               'buffer_input_cb', '',
                                               'buffer_close_cb', '')
    if weetris['buffer']:
        weechat.buffer_set(weetris['buffer'],
                           'type', 'free')
        weechat.buffer_set(weetris['buffer'],
                           'title', BUFFER_TITLE)
        for key, action in keys.items():
            weechat.buffer_set(weetris['buffer'],
                               'key_bind_%s' % key,
                               '/weetris %s' % action)
        new_game()
        weechat.buffer_set(weetris['buffer'], 'display', '1')


def run_action(action):
    """Run an action, when a key is pressed on weetris buffer."""
    if action == 'rotate':
        new_form = rotation(weetris['piece_form'])
        if is_possible(weetris['piece_x'], weetris['piece_y'], new_form):
            weetris['piece_form'] = new_form
            display_all()
    elif action == 'left':
        if is_possible(weetris['piece_x'] - 1, weetris['piece_y'],
                       weetris['piece_form']):
            weetris['piece_x'] -= 1
            display_all()
    elif action == 'right':
        if is_possible(weetris['piece_x'] + 1, weetris['piece_y'],
                       weetris['piece_form']):
            weetris['piece_x'] += 1
            display_all()
    elif action == 'down':
        if is_possible(weetris['piece_x'], weetris['piece_y'] + 1,
                       weetris['piece_form']):
            weetris['piece_y'] += 1
        else:
            end_of_piece()
        display_all()
    elif action == 'bottom':
        while is_possible(weetris['piece_x'], weetris['piece_y'] + 1,
                          weetris['piece_form']):
            weetris['piece_y'] += 1
        end_of_piece()
        display_all()


def weetris_cmd_cb(data, buf, args):
    """Callback for command /weetris."""
    if weetris['buffer']:
        weechat.buffer_set(weetris['buffer'], 'display', '1')
    else:
        weetris_init()

    if args == 'new_game':
        new_game()
    elif args == 'pause':
        if weetris['playing']:
            weetris['paused'] = not weetris['paused']
            display_all()
    else:
        if weetris['playing'] and not weetris['paused']:
            key_slow_down = weetris_settings['key_down_slow'] == 'on'
            actions = {
                'up': 'rotate',
                'left': 'left',
                'right': 'right',
                'down': 'down' if key_slow_down else 'bottom',
                'bottom': 'bottom' if key_slow_down else 'down',
            }
            action = actions.get(args)
            if action:
                run_action(action)
    return weechat.WEECHAT_RC_OK


def weetris_timer_cb(data, remaining_calls):
    """Weetris timer callback."""
    if weetris['buffer'] and weetris['playing'] and not weetris['paused']:
        if is_possible(weetris['piece_x'], weetris['piece_y'] + 1,
                       weetris['piece_form']):
            weetris['piece_y'] += 1
        else:
            end_of_piece()
        display_all()
    return weechat.WEECHAT_RC_OK


def get_best(name, default=0):
    """Get the best level/lines."""
    value = weechat.config_get_plugin(name)
    return int(value) if value else default


def set_best(name, value):
    """Set the best level/lines."""
    weechat.config_set_plugin(name, str(value))


def main():
    """Main function."""
    if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION,
                        SCRIPT_LICENSE, SCRIPT_DESC, '', ''):
        term_colors = int(weechat.info_get('term_colors', '') or '8')
        weetris['256colors'] = term_colors >= 256

        # set default settings
        version = weechat.info_get('version_number', '') or 0
        for option, value in weetris_settings_default.items():
            if weechat.config_is_set_plugin(option):
                weetris_settings[option] = weechat.config_get_plugin(option)
            else:
                weechat.config_set_plugin(option, value[0])
                weetris_settings[option] = value[0]
            if int(version) >= 0x00030500:
                weechat.config_set_desc_plugin(
                    option,
                    '%s (default: "%s")' % (value[1], value[0]))

        # detect config changes
        weechat.hook_config('plugins.var.python.%s.*' % SCRIPT_NAME,
                            'weetris_config_cb', '')

        # command /weetris
        weechat.hook_command('weetris', 'Run WeeTris', '',
                             'Keys:\n'
                             '   arrow up: rotate current piece\n'
                             ' arrow left: move piece to the left\n'
                             'arrow right: move piece to the right\n'
                             ' arrow down: increase speed of the piece\n'
                             '  ctrl+down: move pieve to the bottom\n'
                             '      alt+n: restart the game\n'
                             '      alt+p: pause current game',
                             '', 'weetris_cmd_cb', '')

        # initialization
        if weechat.buffer_search('python', 'weetris'):
            weetris_init()
        weetris['best_level'] = get_best('max_level', 1)
        weetris['best_lines'] = get_best('max_lines', 0)


if __name__ == '__main__' and import_ok:
    main()



"""
    wtwitter.py

    author: Jimmy Zelinskie <jimmyzelinskie@gmail.com>
      desc: Sends your latest tweet to the current buffer
     usage:
       /set plugins.var.python.wtwitter.twitter_handle yourusername
       /wtwitter
   license: GPL3

   history:

       0.2 - rename script to wtwitter.py (twitter is a python module)
       Sebastien Helleu <flashcode@flashtux.org>

       0.1 - initial script
       Jimmy Zelinskie <jimmyzelinskie@gmail.com>

       0.0 - forked from lastfm.py by
       Adam Saponara <saponara TA gmail TOD com>

"""

import weechat
import feedparser

weechat.register("wtwitter", "Jimmy Zelinskie", "0.2", "GPL3", "Sends your latest tweet to the current buffer", "", "")

defaults = {
        "twitter_handle" : "nobody",
        "command" : "/me last tweeted: %s"
}

cmd_hook_process = ""
cmd_buffer       = ""
cmd_stdout       = ""
cmd_stderr       = ""

for k, v in defaults.iteritems():
        if not weechat.config_is_set_plugin(k):
                weechat.config_set_plugin(k, v)

def twitter_cmd(data, buffer, args):
        global cmd_hook_process, cmd_buffer, cmd_stdout, cmd_stderr
        if cmd_hook_process != "":
                weechat.prnt(buffer, "Twitter is already running!")
                return weechat.WEECHAT_RC_OK
        cmd_buffer = buffer
        cmd_stdout = ""
        cmd_stderr = ""
        python2_bin = weechat.info_get("python2_bin", "") or "python"
        cmd_hook_process = weechat.hook_process(
                python2_bin + " -c \"\n"
                "import sys, feedparser\n"
                "feed = None\n"
                "feed = feedparser.parse('http://api.twitter.com/1/statuses/user_timeline.rss?screen_name=%(username)s')\n"
                "if not feed or feed.bozo:\n"
                "	print >>sys.stderr, 'Could not fetch Twitter RSS feed.',\n"
                "elif not 'items' in feed or len(feed['items']) < 1:\n"
                "	print >>sys.stderr, 'No tweets found in Twitter RSS feed.',\n"
                "else:\n"
                "	print '@'+feed['items'][0]['title'].replace(u'\u2013', '-').encode('utf-8', 'replace'),\n"
                "\"" % {"username" : weechat.config_get_plugin('twitter_handle')},
                10000, "twitter_cb", "")
        return weechat.WEECHAT_RC_OK

def twitter_cb(data, command, rc, stdout, stderr):
        global cmd_hook_process, cmd_buffer, cmd_stdout, cmd_stderr
        cmd_stdout += stdout
        cmd_stderr += stderr
        if int(rc) >= 0:
                if cmd_stderr != "":
                        weechat.prnt(cmd_buffer, "%s" % cmd_stderr)
                if cmd_stdout != "":
                        weechat.command(cmd_buffer, weechat.config_get_plugin("command") % cmd_stdout)
                cmd_hook_process = ""
        return weechat.WEECHAT_RC_OK

hook = weechat.hook_command(
        "wtwitter",
        "Sends your latest tweet to the current buffer. Before using /wtwitter, set your twitter handle like this:\n\n"
        "    /set plugins.var.python.wtwitter.twitter_handle yourhandle\n\n"
        "You can also customize the command that will be sent to the buffer like this:\n\n"
        "    /set plugins.var.python.wtwitter.command I last tweeted %s\n",
        "", "", "", "twitter_cmd", "")



# -*- coding: utf-8 -*-
#
# Copyright (c) 2017 Manu Koell <manu@koell.li>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# 2017-11-03: fix script/issue #236
#       v0.2: add "%h" variable in option 'file'
# 2018-10-23: fix script/issue #297
#       v0.3: make script python 3 compatible

from __future__ import print_function

import os
import re

from fnmatch import fnmatch

try:
    import weechat as w

except Exception:
    print("This script must be run under WeeChat.")
    print("Get WeeChat now at: http://www.weechat.org/")
    quit()

NAME        = "autoconf"
AUTHOR      = "Manu Koell <manu@koell.li>"
VERSION     = "0.3"
LICENSE     = "GPL3"
DESCRIPTION = "auto save/load changed options in a ~/.weerc file, useful to share dotfiles with"

EXCLUDES = [
    '*.nicks',
    '*.username', '*.sasl_username',
    '*.password', '*.sasl_password',
    'irc.server.*.autoconnect',
    'irc.server.*.autojoin'
]

SETTINGS = {
    'autosave': ('on', 'auto save config on quit'),
    'autoload': ('on', 'auto load config on start'),
    'ignore': (
        ','.join(EXCLUDES),
        'comma separated list of patterns to exclude'),
    'file': ('%h/.weerc', 'config file location ("%h" will be replaced by WeeChat home, "~/.weechat" by default)')
}

def cstrip(text):
    """strip color codes"""

    return w.string_remove_color(text, '')

def get_config(args):
    """get path to config file"""

    try:
        conf = args[1]
    except Exception:
        conf = w.config_get_plugin('file').replace("%h",w.info_get("weechat_dir", ""))
    return os.path.expanduser(conf)

def load_conf(args):
    """send config to fifo pipe"""

    fifo = w.info_get('fifo_filename', '')
    conf = get_config(args)

    if os.path.isfile(conf):
        w.command('', '/exec -sh -norc cat | grep */set %s > %s' % (conf, fifo))

def save_conf(args):
    """match options and save to config file"""

    try:
        f = open(get_config(args), 'w+')

    except Exception as e:
        w.prnt('', '%sError: %s' % (w.prefix('error'), e))

        return w.WEECHAT_RC_ERROR

    header = [
        '#',
        '# WeeChat %s (compiled on %s)' % (w.info_get('version', ''), w.info_get('date', '')),
        '#',
        '# Use /autoconf load or cat this file to the FIFO pipe.',
        '#',
        '# For more info, see https://weechat.org/scripts/source/autoconf.py.html',
        '#',
        ''
    ]

    for ln in header:
        f.write('%s\n' % ln)

    w.command('', '/buffer clear')
    w.command('', '/set diff')

    infolist = w.infolist_get('buffer_lines', '', '')

    while w.infolist_next(infolist):
        message = cstrip(w.infolist_string(infolist, 'message'))
        ignore = w.config_get_plugin('ignore').split(',')
        option = re.match(RE['option'], message)

        if option:
            if not any(fnmatch(option.group(1), p.strip()) for p in ignore):
                f.write('*/set %s %s\n' % (option.group(1), option.group(2)))

    f.close()

    w.infolist_free(infolist)

def autoconf_cb(data, buffer, args):
    """the /autoconf command"""

    args = args.split()

    if 'save' in args:
        save_conf(args)

    elif 'load' in args:
        load_conf(args)

    else:
        # show help message
        w.command('', '/help ' + NAME)

    return w.WEECHAT_RC_OK

def quit_cb(data, signal, signal_data):
    """save config on quit"""

    save_conf(None)

    return w.WEECHAT_RC_OK

if __name__ == '__main__':
    if w.register(NAME, AUTHOR, VERSION, LICENSE, DESCRIPTION, "", ""):
        w.hook_command(NAME, DESCRIPTION, 'save [path] || load [path]', '', 'save || load', 'autoconf_cb', '')
        default_txt = w.gettext("default: ")            # check if string is translated
        RE = {
            'option': re.compile('\s*(.*) = (.*)  \(%s' % default_txt)
        }

        # set default config
        for option, value in SETTINGS.items():
            if not w.config_is_set_plugin(option):
                w.config_set_plugin(option, value[0])
                w.config_set_desc_plugin(option, '%s  (default: "%s")' % (value[1], value[0]))

        if 'on' in w.config_get_plugin('autoload'):
            load_conf(None)

        if 'on' in w.config_get_plugin('autosave'):
            w.hook_signal('quit', 'quit_cb', '')



# ===============================================================
# crypt.py (c) 2008-2012 by Nicolai Lissner <nlissne@linux01.org>
# ===============================================================
SCRIPT_NAME    = "crypt"
SCRIPT_AUTHOR  = "Nicolai Lissner <nlissne@linux01.org>"
SCRIPT_VERSION = "1.4.4"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC    = "encrypt/decrypt PRIVMSGs using a pre-shared key and openssl"

#
# This plugin uses openssl to encrypt/decrypt messages you send
# or receive with weechat. Due to the very simple method
# used it should be easy to port this plugin to any other
# IRC client.
#
# The default encryption algorithm is blowfish, but you can
# change it to any other cipher your openssl offers.
# Read output of "openssl -h" to find out which ciphers are
# supported (also depends on your kernel!)
#
# To activate encryption for a given user just
# put a file called "cryptkey.username" into
# your weechat-directory, containing the passphrase
# to use for encryption/decryption
#
# you can add 'encryption' to
# weechat.bar.status.items to have an indication
# that the message you are going to send is encrypted
# (i.e. a keyfile exists)
#
# You can activate encryption on irc-channels, too,
# just use cryptkey.#channelname as keyfile then.
#
# example: if you have exchanged a secret key with me,
# you would put it in a file called
# cryptkey.blackpenguin in your weechat_dir
#
# Of course, you need to share this keyfile with the
# remote side in another secure way (i.e. sending
# pgp-encrypted mail)

import weechat, string, os, subprocess, re

script_options = {
    "message_indicator" : "(enc) ",
    "statusbar_indicator" : "(encrypted) ",
    "cipher" : "blowfish",
}

def decrypt(data, msgtype, servername, args):
  hostmask, chanmsg = string.split(args, "PRIVMSG ", 1)
  channelname, message = string.split(chanmsg, " :", 1)
  if re.match(r'^\[\d{2}:\d{2}:\d{2}]\s', message):
    timestamp = message[:11]
    message = message[11:]
  else:
    timestamp = ''
  if channelname[0] == "#":
    username=channelname
  else:
    username, rest = string.split(hostmask, "!", 1)
    username = username[1:]
  if os.path.exists(weechat_dir + "/cryptkey." + username):
    p = subprocess.Popen(["openssl", "enc", "-d", "-a", "-" + weechat.config_get_plugin("cipher"), "-pass" ,"file:" + weechat_dir + "/cryptkey." + username], bufsize=4096, stdin=PIPE, stdout=PIPE, stderr=PIPE, close_fds=True)
    p.stdin.write("U2FsdGVkX1" + message.replace("|","\n"))
    p.stdin.close()
    decrypted = p.stdout.read()
    p.stdout.close()
    if decrypted == "":
      return args
    decrypted = ''.join(c for c in decrypted if ord(c) > 31 or ord(c) == 9 or ord(c) == 2 or ord(c) == 3 or ord(c) == 15)
    return hostmask + "PRIVMSG " + channelname + " :" + chr(3) + "04" + weechat.config_get_plugin("message_indicator") + chr(15) + timestamp + decrypted
  else:
    return args

def encrypt(data, msgtype, servername, args):
  pre, message = string.split(args, ":", 1)
  prestr=pre.split(" ")
  username=prestr[-2]
  if os.path.exists(weechat_dir + "/cryptkey." + username):
    p = subprocess.Popen(["openssl", "enc", "-a", "-" + weechat.config_get_plugin("cipher"), "-pass" ,"file:" + weechat_dir + "/cryptkey." + username], bufsize=4096, stdin=PIPE, stdout=PIPE, stderr=PIPE, close_fds=True)
    p.stdin.write(message)
    p.stdin.close()
    encrypted = p.stdout.read()
    p.stdout.close()
    encrypted = encrypted.replace("\n","|")
    if len(encrypted) > 400:
      splitmsg=string.split(message," ")
      cutpoint=len(splitmsg)/2
      p = subprocess.Popen(["openssl", "enc", "-a", "-" + weechat.config_get_plugin("cipher"), "-pass" ,"file:" + weechat_dir + "/cryptkey." + username], bufsize=4096, stdin=PIPE, stdout=PIPE, stderr=PIPE, close_fds=True)
      p.stdin.write(string.join(splitmsg[:cutpoint]," ") + "\n")
      p.stdin.close()
      encrypted = p.stdout.read()
      p.stdout.close()
      encrypted = encrypted.replace("\n","|")
      p = subprocess.Popen(["openssl", "enc", "-a", "-" + weechat.config_get_plugin("cipher"), "-pass" ,"file:" + weechat_dir + "/cryptkey." + username], bufsize=4096, stdin=PIPE, stdout=PIPE, stderr=PIPE, close_fds=True)
      p.stdin.write( string.join(splitmsg[cutpoint:]," ") )
      p.stdin.close()
      encrypted2 = p.stdout.read()
      p.stdout.close()
      encrypted2 = encrypted2.replace("\n","|")
      encrypted = encrypted + "\n" + pre + ":" + encrypted2[10:]
    return pre + ":" + encrypted[10:]
  else:
    return args

def update_encryption_status(data, signal, signal_data):
    buffer = signal_data
    weechat.bar_item_update('encryption')
    return weechat.WEECHAT_RC_OK

def encryption_statusbar(data, item, window):
    if window:
      buf = weechat.window_get_pointer(window, 'buffer')
    else:
      buf = weechat.current_buffer()
    if os.path.exists(weechat_dir + "/cryptkey." + weechat.buffer_get_string(buf, "short_name")):
      return weechat.config_get_plugin("statusbar_indicator")
    else:
      return ""


# for subprocess.Popen call
PIPE=-1

# register plugin
if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "", "UTF-8"):
    weechat_dir = weechat.info_get("weechat_dir","")
    version = weechat.info_get("version_number", "") or 0
    if int(version) < 0x00030000:
      weechat.prnt("", "%s%s: WeeChat 0.3.0 is required for this script."
              % (weechat.prefix("error"), SCRIPT_NAME))
    else:
      weechat.bar_item_new('encryption', 'encryption_statusbar', '')
      for option, default_value in script_options.iteritems():
          if not weechat.config_is_set_plugin(option):
                  weechat.config_set_plugin(option, default_value)
      # register the modifiers
      weechat.hook_modifier("irc_in_privmsg", "decrypt", "")
      weechat.hook_modifier("irc_out_privmsg", "encrypt", "")
      weechat.hook_signal("buffer_switch","update_encryption_status","")



# -*- coding: utf-8 -*-
#
# Copyright (c) 2012-2019 by nils_2 <weechatter@arcor.de>
#
# add a plain text or evaluated content to item bar
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# 2019-05-23: FlashCode, (freenode.#weechat)
#       0.9 : fix eval_expression() for split windows
#
# 2018-08-18: nils_2, (freenode.#weechat)
#       0.8 : add new option "interval"
#
# 2017-08-23: nils_2, (freenode.#weechat)
#     0.7.1 : improve /help text
#
# 2017-08-19: nils_2, (freenode.#weechat)
#       0.7 : add type "!all", internal changes
#
# 2016-12-12: nils_2, (freenode.#weechat)
#       0.6 : fix problem with multiple windows (reported by Ram-Z)
#
# 2016-09-15: nils_2, (freenode.#weechat)
#       0.5 : add /help text (suggested by gb)
#
# 2014-05-19: nils_2, (freenode.#weechat)
#       0.4 : evaluate content of item (suggested by FlashCode)
#
# 2013-06-27: nils_2, (freenode.#weechat)
#       0.3 : fix: bug with root bar
#
# 2013-01-25: nils_2, (freenode.#weechat)
#       0.2 : make script compatible with Python 3.x
#
# 2012-12-23: nils_2, (freenode.#weechat)
#       0.1 : initial release
#
# requires: WeeChat version 0.3.0
#
# Development is currently hosted at
# https://github.com/weechatter/weechat-scripts

# TODO
# plugins.var.python.text_item.<item_name>.enabled
# plugins.var.python.text_item.<item_name>.type
# plugins.var.python.text_item.<item_name>.signal
# plugins.var.python.text_item.<item_name>.text
# plugins.var.python.text_item.<item_name>.interval

try:
    import weechat,re

except Exception:
    print("This script must be run under WeeChat.")
    print("Get WeeChat now at: http://www.weechat.org/")
    quit()

SCRIPT_NAME     = "text_item"
SCRIPT_AUTHOR   = "nils_2 <weechatter@arcor.de>"
SCRIPT_VERSION  = "0.9"
SCRIPT_LICENSE  = "GPL"
SCRIPT_DESC     = "add a plain text or evaluated content to item bar"

# regexp to match ${color} tags
regex_color=re.compile('\$\{([^\{\}]+)\}')

hooks = {}
TIMER = None

settings = {
        'interval': ('0', 'How often (in seconds) to force an update of all items. 0 means deactivated'),
}
# ================================[ hooks ]===============================
def add_hook(signal, item):
    global hooks
    # signal already exists?
    if signal in hooks:
        return
    hooks[item] = weechat.hook_signal(signal, "bar_item_update_cb", "")

def unhook(hook):
    global hooks
    if hook in hooks:
        weechat.unhook(hooks[hook])
        del hooks[hook]

def toggle_refresh_cb(pointer, name, value):
    option_name = name[len('plugins.var.python.' + SCRIPT_NAME + '.'):]      # get optionname

    # check for timer hook
    if name.endswith(".interval"):
        set_timer()
        return weechat.WEECHAT_RC_OK

    # option was removed? remove bar_item from struct
    if not weechat.config_get_plugin(option_name):
        ptr_bar = weechat.bar_item_search(option_name)
        if ptr_bar:
            weechat.bar_item_remove(ptr_bar)
        return weechat.WEECHAT_RC_OK

    # check if option is new or changed
    if not weechat.bar_item_search(option_name):
        weechat.bar_item_new(option_name,'update_item',option_name)

    weechat.bar_item_update(option_name)
    return weechat.WEECHAT_RC_OK

def set_timer():
    # Update timer hook with new interval. 0 means deactivated
    global TIMER
    if TIMER:
        weechat.unhook(TIMER)
    if int(weechat.config_get_plugin('interval')) >= 1:
        TIMER = weechat.hook_timer(int(weechat.config_get_plugin('interval')) * 1000,0, 0, "timer_dummy_cb", '')

def timer_dummy_cb(data, remaining_calls):
    # hook_timer() has two arguments, hook_signal() needs three arguments
    bar_item_update_cb("","","")
    return weechat.WEECHAT_RC_OK

# ================================[ items ]===============================
def create_bar_items():
    ptr_infolist_option = weechat.infolist_get('option','','plugins.var.python.' + SCRIPT_NAME + '.*')

    if not ptr_infolist_option:
        return

    while weechat.infolist_next(ptr_infolist_option):
        option_full_name = weechat.infolist_string(ptr_infolist_option, 'full_name')
        option_name = option_full_name[len('plugins.var.python.' + SCRIPT_NAME + '.'):]      # get optionname

        if weechat.bar_item_search(option_name):
            weechat.bar_item_update(option_name)
        else:
            weechat.bar_item_new(option_name,'update_item',option_name)
        weechat.bar_item_update(option_name)

    weechat.infolist_free(ptr_infolist_option)

def update_item (data, item, window):
    if not data:
        return ""

    # window empty? root bar!
    if not window:
        window = weechat.current_window()

    value = weechat.config_get_plugin(data)
    if not value:
        return ""

    value = check_buffer_type(window, data, value)

    return substitute_colors(value,window)

# update item from weechat.hook_signal()
def bar_item_update_cb(signal, callback, callback_data):
    ptr_infolist_option = weechat.infolist_get('option','','plugins.var.python.' + SCRIPT_NAME + '.*')

    if not ptr_infolist_option:
        return

    while weechat.infolist_next(ptr_infolist_option):
        option_full_name = weechat.infolist_string(ptr_infolist_option, 'full_name')
        option_name = option_full_name[len('plugins.var.python.' + SCRIPT_NAME + '.'):]      # get optionname
        if option_name == "interval":
            continue

        # check if item exists in a bar and if we have a hook for it
        if weechat.bar_item_search(option_name) and option_name in hooks:
            weechat.bar_item_update(option_name)

    weechat.infolist_free(ptr_infolist_option)

    return weechat.WEECHAT_RC_OK


# ================================[ subroutines ]===============================
def substitute_colors(text,window):
    if int(version) >= 0x00040200:
        bufpointer = weechat.window_get_pointer(window,"buffer")
        return weechat.string_eval_expression(text, {"window": window, "buffer": bufpointer}, {}, {})
    # substitute colors in output
    return re.sub(regex_color, lambda match: weechat.color(match.group(1)), text)

def check_buffer_type(window, data, value):
    bufpointer = weechat.window_get_pointer(window,"buffer")
    if bufpointer == "":
        return ""

    value = value.split(' ', 1)
    if len(value) <= 1:
        return ""

    # format is : buffer_type (channel,server,private,all) | signal (e.g: buffer_switch)
    channel_type_and_signal = value[0]
    if channel_type_and_signal.find('|') >= 0:
        channel_type = channel_type_and_signal[0:channel_type_and_signal.find("|")]
        signal_type = channel_type_and_signal[channel_type_and_signal.find("|")+1:]
        unhook(data)
        add_hook(signal_type, data)
    else:
        channel_type = value[0]

    value = value[1]

    if channel_type == 'all' or weechat.buffer_get_string(bufpointer,'localvar_type') == channel_type:
        return value
    if channel_type == '!all':
        a = ["channel","server","private"]
        if weechat.buffer_get_string(bufpointer,'localvar_type') in a:
            return value
    return ""

# ================================[ main ]===============================
if __name__ == "__main__":
    if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC,'',''):
            weechat.hook_command(SCRIPT_NAME,SCRIPT_DESC,
                        '',
                        'How to use:\n'
                        '===========\n'
                        'Template:\n'
                        '/set plugins.var.python.text_item.<item_name> <type>|<signal> <${color:name/number}><text>\n\n'
                        '   type : channel, server, private, all (all kind of buffers e.g. /color, /fset...) and !all (channel, server and private buffer)\n'
                        '   (see: /buffer localvar)\n\n'
                        '   signal (eg.): buffer_switch, buffer_closing, print, mouse_enabled\n'
                        '   (for a list of all possible signals, see API doc weechat_hook_signal())\n'
                        '\n'
                        'You can activate a timer hook() to force an upgrade of all items in a given period of time, for example using an item that have to be\n'
                        'updated every second (e.g. watch)\n'
                        '\n'
                        'Examples:\n'
                        'creates an option for a text item named "nick_text". The item will be created for "channel" buffers. '
                        'The text displayed in the status-bar is "Nicks:" (yellow colored!):\n'
                        '   /set plugins.var.python.text_item.nick_text "channel ${color:yellow}Nicks:"\n\n'
                        'now you have to add the item "nick_text" to the bar.items (use auto-completion or iset.pl!)\n'
                        '   /set weechat.bar.status.items nick_text\n\n'
                        'creates an option to display the terminal width and height in an item bar. item will be updated on signal "signal_sigwinch":\n'
                        '   /set plugins.var.python.text_item.dimension "all|signal_sigwinch width: ${info:term_width} height: ${info:term_height}"\n'
                        'creates an option to display the status from "/filter toggle" and "/filter toggle @" command, item name is "filter_item":\n'
                        '   /set plugins.var.python.text_item.filter_item "!all|*filters* ${if:${info:filters_enabled}==1?${color:yellow}F:${color:243}F}${if:${buffer.filter}==1?${color:yellow}@:${color:243}@}"\n',
                        '',
                        '',
                        '')
            version = weechat.info_get("version_number", "") or 0

            for option, default_desc in settings.items():
                if not weechat.config_is_set_plugin(option):
                    weechat.config_set_plugin(option, default_desc[0])
                if int(version) >= 0x00030500:
                    weechat.config_set_desc_plugin(option, default_desc[1])

            set_timer()
            create_bar_items()

            weechat.hook_config( 'plugins.var.python.' + SCRIPT_NAME + '.*', 'toggle_refresh_cb', '' )



# ===============================================================

# Copyright (C) 2014 by David R. Andersen and Tycho Andersen

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Development repository at: https://github.com/rxcomm/weeText/

SCRIPT_NAME    = "weetext"
SCRIPT_AUTHOR  = "David R. Andersen <k0rx@RXcomm.net>, Tycho Andersen <tycho@tycho.ws>"
SCRIPT_VERSION = "0.1.2"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC    = "SMS Text Messaging script for Weechat using Google Voice"

"""
This script implements chatting via text message with Weechat.

Email and password should be configured (either by editing the script
itself before loading or adding options to plugins.conf). For using
secure passwords, see the weechat /secure command.

To initiate a text message session with someone new, that isn't currently
in your weeText buffer list, in the weeText buffer type the command:

text <10 digit phone number>

This will pop open a new buffer.

You can also send text messages to multiple numbers. The syntax is (from
the weeText buffer):

multi <number1>,<number2>,...

This will pop open a new multi-text buffer.

I've also added optional symmetric-key encryption using OpenSSL. This is
essentially a wholesale copy of the encrypt() and decrypt() methods from
the weechat crypt.py script. Thanks to the authors for that!

Todo:
1. Add buffer for texting multiple parties at the same time.

"""

import weechat
import sys
import os
import glob
import re
import cPickle
import subprocess
import random
import string
from googlevoice import Voice
from googlevoice.util import input
from BeautifulSoup import BeautifulSoup, BeautifulStoneSoup, SoupStrainer

script_options = {
    "email" : "", # GV email address
    "passwd" : "", # GV password - can use /secure
    "poll_interval" : "120", # poll interval for receiving messages (sec)
    "encrypt_sms" : "True",
    "key_dir" : "/cryptkey",
    "cipher" : "aes-256-cbc",
    "message_indicator" : "(enc) ",
}

conversation_map = {}
number_map = {}
conv = ''

class Conversation(object):
    def __init__(self, conv_id, number, messages):
        self.conv_id = conv_id
        self.number = number
        self.messages = messages

    def new_messages(self, other):
        assert len(self.messages) <= len(other.messages)
        return other.messages[len(self.messages):]

    def __iter__(self):
        return iter(reversed(self.messages))

def renderConversations(unused, command, return_code, out, err):
    global conversation_map
    global conv

    if return_code == weechat.WEECHAT_HOOK_PROCESS_ERROR:
        weechat.prnt("", "Error with command '%s'" % command)
        return weechat.WEECHAT_RC_OK
    if return_code > 0:
        weechat.prnt("", "return_code = %d" % return_code)
    if out != '':
        conv += out
        if return_code == weechat.WEECHAT_HOOK_PROCESS_RUNNING:
            weechat.prnt('', 'getting more data')
            return weechat.WEECHAT_RC_OK
    if err != "":
        weechat.prnt("", "stderr: %s" % err)
        return weechat.WEECHAT_RC_OK

    try:
        conversations = reversed(cPickle.loads(conv))
    except EOFError:
        weechat.prnt('', 'wtrecv returned garbage')
        return weechat.WEECHAT_RC_OK

    for conversation in conversations:
        if not conversation.conv_id in conversation_map:
            conversation_map[conversation.conv_id] = conversation
            msgs = conversation.messages
        else:
            old = conversation_map[conversation.conv_id]
            conversation_map[conversation.conv_id] = conversation
            msgs = old.new_messages(conversation)
        for msg in msgs:
            if not conversation.number in number_map and msg['from'] != 'Me:':
                number_map[conversation.number] = msg['from']
        for msg in msgs:
            if conversation.number in number_map:
                buf = weechat.buffer_search('python', number_map[conversation.number][:-1])
                if not buf:
                    buf = weechat.buffer_new(number_map[conversation.number][:-1],
                                             "textOut", "", "buffer_close_cb", "")
            else:
                buf = weechat.buffer_search('python', 'Me')
                if not buf:
                    buf = weechat.buffer_new('Me', "textOut", "", "buffer_close_cb", "")
            if weechat.config_get_plugin('encrypt_sms') == 'True':
                msg['text'] = decrypt(msg['text'], buf)
            nick = msg['from'][:-1].strip()
            tags = 'notify_private,nick_' + msg['from'][:-1].strip()
            tags += ',log1,prefix_nick_' + weechat.info_get('irc_nick_color_name', nick)
            nick = msg['from'][:-1].strip()
            weechat.prnt_date_tags(buf, 0, tags, '\x03' + weechat.info_get('irc_nick_color', nick)
                                   + nick + '\t' + msg['text'])
    conv = ''
    callGV()
    return weechat.WEECHAT_RC_OK

def textOut(data, buf, input_data):
    global number_map
    number = None
    for num, dest in number_map.iteritems():
        if dest[:-1] == weechat.buffer_get_string(buf, 'name'):
            number = num[2:]
    if not number:
        number = weechat.buffer_get_string(buf, 'name')[2:]
    if weechat.config_get_plugin('encrypt_sms') == 'True':
        input_data = encrypt(input_data, buf)
    msg_id = ''.join(random.choice(string.lowercase) for x in range(4))
    callGV(buf=buf, number=number, input_data=input_data, msg_id=msg_id, send=True)
    return weechat.WEECHAT_RC_OK

def multiText(data, buf, input_data):
    global number_map
    numbers = data.split(',')
    if weechat.config_get_plugin('encrypt_sms') == 'True':
        input_data = encrypt(input_data, buf)
    for number in numbers:
        msg_id = ''.join(random.choice(string.lowercase) for x in range(4))
        callGV(buf=buf, number=number, input_data=input_data, msg_id=msg_id, send=True)
    return weechat.WEECHAT_RC_OK

def sentCB(buf_name, command, return_code, out, err):
    if return_code == weechat.WEECHAT_HOOK_PROCESS_ERROR:
        weechat.prnt("", "Error with command '%s'" % command)
        return weechat.WEECHAT_RC_OK
    if return_code > 0:
        weechat.prnt("", "return_code = %d" % return_code)
    if out != "":
        tags = 'notify_message'
        weechat.prnt_date_tags(weechat.buffer_search('python', buf_name), 0, tags, out)
    if err != "":
        weechat.prnt("", "stderr: %s" % err)
    return weechat.WEECHAT_RC_OK

def gvOut(data, buf, input_data):
    if input_data[:4] == 'text' and buf == weechat.buffer_search('python', 'weeText'):
        buffer = weechat.buffer_new("+1"+input_data[5:], "textOut", "", "buffer_close_cb", "")
    if input_data[:5] == 'multi' and buf == weechat.buffer_search('python', 'weeText'):
        num_list = input_data[6:].split(',')
        nums = ''
        for num in num_list:
            nums += '+' + num[-4:]
        nums = nums[1:]
        buffer = weechat.buffer_new('m:' + nums, "multiText", input_data[6:], "buffer_close_cb", "")
    return weechat.WEECHAT_RC_OK

def buffer_input_cb(data, buf, input_data):
    # ...
    return weechat.WEECHAT_RC_OK

def buffer_close_cb(data, buf):
    return weechat.WEECHAT_RC_OK

def encrypt(message, buf):
  username=weechat.buffer_get_string(buf, 'name')
  if os.path.exists(weechat_dir + key_dir + "/cryptkey." + username):
    p = subprocess.Popen(["openssl", "enc", "-a", "-" + weechat.config_get_plugin("cipher"),
                          "-pass" ,"file:" + weechat_dir + key_dir + "/cryptkey." + username],
                          bufsize=4096, stdin=PIPE, stdout=PIPE, stderr=PIPE, close_fds=True)
    p.stdin.write(message)
    p.stdin.close()
    encrypted = p.stdout.read()
    p.stdout.close()
    encrypted = encrypted.replace("\n","|")
    return encrypted[10:]
  else:
    return message

def decrypt(message, buf):
  username=weechat.buffer_get_string(buf, 'name')
  if os.path.exists(weechat_dir + key_dir + "/cryptkey." + username):
    p = subprocess.Popen(["openssl", "enc", "-d", "-a", "-" + weechat.config_get_plugin("cipher"),
                          "-pass" ,"file:" + weechat_dir + key_dir + "/cryptkey." + username],
                          bufsize=4096, stdin=PIPE, stdout=PIPE, stderr=PIPE, close_fds=True)
    p.stdin.write("U2FsdGVkX1" + message.replace("|","\n"))
    p.stdin.close()
    decrypted = p.stdout.read()
    p.stdout.close()
    if decrypted == "":
      return message
    decrypted = ''.join(c for c in decrypted if ord(c) > 31 or ord(c) == 9 or ord(c) == 2
                or ord(c) == 3 or ord(c) == 15)
    return '\x19' + weechat.color('lightred') + weechat.config_get_plugin("message_indicator") + '\x1C' + decrypted
  else:
    return message

def update_encryption_status(data, signal, signal_data):
    buffer = signal_data
    weechat.bar_item_update('encryption')
    return weechat.WEECHAT_RC_OK

def encryption_statusbar(data, item, window):
    if window:
      buf = weechat.window_get_pointer(window, 'buffer')
    else:
      buf = weechat.current_buffer()
    if os.path.exists(weechat_dir + key_dir + "/cryptkey." + weechat.buffer_get_string(buf, "short_name")):
      return weechat.config_get_plugin("statusbar_indicator")
    else:
      return ""

def checkWTrecv(*args):
    tmp = os.popen('ps -Af').read()
    if not tmp.count('wtrecv'):
        callGV()
    return weechat.WEECHAT_RC_OK


def callGV(buf=None, number=None, input_data=None, msg_id=None, send=False):
    if send:
        send_hook = weechat.hook_process_hashtable(weechat_dir + '/python/wtsend.py',
                    { 'stdin': '' }, 0, 'sentCB', weechat.buffer_get_string(buf, 'name'))
        proc_data = email + '\n' + passwd + '\n' + number + '\n' +\
                    input_data + '\n' + msg_id + '\n'
        weechat.hook_set(send_hook, 'stdin', proc_data)
    else:
        proc_data = email + '\n' + passwd + '\n' +\
                    weechat.config_get_plugin('poll_interval') + '\n'
        recv_hook = weechat.hook_process_hashtable(weechat_dir + '/python/wtrecv.py',
                { 'stdin': '' }, 0, 'renderConversations', '')
        weechat.hook_set(recv_hook, 'stdin', proc_data)

PIPE=-1

# register plugin
if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "", "UTF-8"):
    buffer = weechat.buffer_new("weeText", "gvOut", "", "buffer_close_cb", "")
    weechat_dir = weechat.info_get("weechat_dir","")
    key_dir = weechat.config_get_plugin("key_dir")
    weechat.bar_item_new('encryption', 'encryption_statusbar', '')
    for option, default_value in script_options.iteritems():
        if not weechat.config_is_set_plugin(option):
            weechat.config_set_plugin(option, default_value)

    # get email/passwd and pass to other script
    email=weechat.config_get_plugin('email')
    passwd = weechat.config_get_plugin('passwd')
    if re.search('sec.*data', passwd):
        passwd=weechat.string_eval_expression(passwd, {}, {}, {})

    # write the helper files
    with open(weechat_dir + '/python/wtrecv.py', 'w') as f:
        f.write("""#!/usr/bin/env python

import sys
import cPickle
import time
import re
import os
import glob
from googlevoice import Voice
from googlevoice.util import input
from BeautifulSoup import BeautifulSoup, BeautifulStoneSoup, SoupStrainer

user_path = os.path.expanduser('~')

class Conversation(object):
    def __init__(self, conv_id, number, messages):
        self.conv_id = conv_id
        self.number = number
        self.messages = messages

    def new_messages(self, other):
        assert len(self.messages) <= len(other.messages)
        return other.messages[len(self.messages):]

    def __iter__(self):
        return iter(reversed(self.messages))

class SMS:

    def getsms(self):
        # We could call voice.sms() directly, but I found this does a rather
        # inefficient parse of things which pegs a CPU core and takes ~50 CPU
        # seconds, while this takes no time at all.
        data = voice.sms.datafunc()
        data = re.search(r'<html><\!\[CDATA\[([^\]]*)', data, re.DOTALL).groups()[0]

        divs = SoupStrainer(['div', 'input'])
        tree = BeautifulSoup(data, parseOnlyThese=divs)

        convos = []
        conversations = tree.findAll("div", attrs={"id" : True},recursive=False)
        for conversation in conversations:
            inputs = SoupStrainer('input')
            tree_inp = BeautifulSoup(str(conversation),parseOnlyThese=inputs)
            phone = tree_inp.find('input', "gc-quickcall-ac")['value']

            smses = []
            msgs = conversation.findAll(attrs={"class" : "gc-message-sms-row"})
            for row in msgs:
                msgitem = {"id" : conversation["id"]}
                spans = row.findAll("span", attrs={"class" : True}, recursive=False)
                for span in spans:
                    cl = span["class"].replace('gc-message-sms-', '')
                    msgitem[cl] = (" ".join(span.findAll(text=True))).strip()
                if msgitem["text"]:
                    msgitem["text"] = BeautifulStoneSoup(msgitem["text"],
                                      convertEntities=BeautifulStoneSoup.HTML_ENTITIES
                                      ).contents[0]
                    msgitem['phone'] = phone
                    smses.append(msgitem)
            convos.append(Conversation(conversation['id'], phone, smses))
        print cPickle.dumps(convos)

if __name__ == '__main__':

    email = sys.stdin.readline().strip()
    passwd = sys.stdin.readline().strip()
    poll_interval = sys.stdin.readline().strip()

    time.sleep(float(poll_interval))

    # create voice instance if no texts are being sent
    while True:
        f = glob.glob(user_path + '/.weechat/.gvlock*')
        if f == []:
            voice = Voice()
            voice.login(email=email, passwd=passwd)
            sms = SMS()
            sms.getsms()
            break
        else:
            time.sleep(1)
""")
    os.chmod(weechat_dir + '/python/wtrecv.py', 0755)

    with open(weechat_dir + '/python/wtsend.py', 'w') as f:
        f.write("""#!/usr/bin/env python

import sys
import os
from googlevoice import Voice
from googlevoice.util import input

# read the credentials, payload, and msg_id from stdin
email = sys.stdin.readline().strip()
passwd = sys.stdin.readline().strip()
number = sys.stdin.readline().strip()
payload = sys.stdin.readline().strip()
msg_id = sys.stdin.readline().strip()

user_path = os.path.expanduser('~')
open(user_path + '/.weechat/.gvlock.' + msg_id, 'a').close()

try:
    voice = Voice()
    voice.login(email, passwd)
    voice.send_sms(number, payload)
    print '<message sent>'
except:
    print '<message NOT sent!>'

os.remove(user_path + '/.weechat/.gvlock.' + msg_id)
""")
    os.chmod(weechat_dir + '/python/wtsend.py', 0755)

    # remove any old .gvlock.* files
    for gvlockfile in glob.glob(weechat_dir + '/.gvlock.*'):
        os.remove(gvlockfile)

    # register the hooks
    weechat.hook_signal("buffer_switch","update_encryption_status","")
    callGV()

    # make sure we are receiving data
    weechat.hook_timer(600000, 0, 0, 'checkWTrecv', '')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Arnaud Renevier <arno@renevier.net>
#
# > updated by kbdkode <kbdkode@protonmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

SCRIPT_NAME    = "autoconnect"
SCRIPT_AUTHOR  = "arno <arno@renevier.net>"
SCRIPT_VERSION = "0.3.3"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC    = "reopens servers and channels opened last time weechat closed"
SCRIPT_COMMAND = "autoconnect"

try:
    import weechat
except:
    print("This script must be run under WeeChat.")
    print("Get WeeChat now at: https://weechat.org/")
    quit()


weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "", "")

quitting = False


def get_autojoin_channels(server):
    channels = weechat.config_string(weechat.config_get("irc.server.%s.autojoin" % (server,)))
    return set(channels.split(',')) if channels else set()


def update_autojoin_channels(server, channels):
    channels = ','.join(channels) if channels else "null"
    weechat.command("", "/mute /set irc.server.%s.autojoin %s" % (server, channels))


def joinpart_cb(data, signal, signal_data):
    server = signal.split(',')[0]
    if weechat.info_get("irc_nick_from_host", signal_data) != weechat.info_get("irc_nick", server):
        # nick which has joined is not our current nick
        return weechat.WEECHAT_RC_OK

    autojoin_channels = get_autojoin_channels(server)

    if signal.endswith("irc_in2_JOIN"):
        weechat.command("", "/mute /set irc.server.%s.autoconnect on" % (server,))

        # get all channels joined (without passphrases)
        chans = [j.split()[0].strip().lstrip(':') for j in signal_data.split(None, 2)[2].split(',')]
        autojoin_channels.add(','.join(chans))

    elif signal.endswith("irc_in2_PART"):
        channel = signal_data.split(' PART ')[1].split()[0]
        autojoin_channels.discard(channel)

    update_autojoin_channels(server, autojoin_channels)
    weechat.command("", "/save irc")
    return weechat.WEECHAT_RC_OK


def disconnect_cb(data, signal, signal_data):
    global quitting
    if not quitting:
        server = signal_data.split(',')[0]

        weechat.command("", "/mute /set irc.server.%s.autoconnect null" % (server,))
        weechat.command("", "/mute /set irc.server.%s.autojoin null" % (server,))

        weechat.command("", "/mute /save irc")
    return weechat.WEECHAT_RC_OK


def quit_cb(data, signal, signal_data):
    global quitting
    quitting = True
    return weechat.WEECHAT_RC_OK


weechat.hook_signal("quit", "quit_cb", "")
weechat.hook_signal("*,irc_in2_join", "joinpart_cb", "")
weechat.hook_signal("*,irc_in2_part", "joinpart_cb", "")
weechat.hook_signal("irc_server_disconnected", "disconnect_cb", "")



# -*- coding: utf-8 -*-
###
# Copyright (c) 2010 by Elián Hanisch <lambdae2@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###

###
#
#   This scripts adds word completion, like irssi's /completion
#   (depends on WeeChat 0.3.1 or newer)
#
#   Commands:
#   * /completion: see /help completion
#
#
#   Settings:
#   * plugins.var.python.completion.replace_values:
#     Completion list, it shouldn't be edited by hand.
#
#
#   History:
#   2019-08-20
#   version 0.3: Ben Harris (benharri)
#   * port for python3
#
#   2010-05-08
#   version 0.2:
#   * complete any word behind the cursor, not just the last one in input line.
#   * change script display name 'completion' to 'cmpl'.
#
#   2010-01-26
#   version 0.1: release
#
###

try:
    import weechat
    WEECHAT_RC_OK = weechat.WEECHAT_RC_OK
    import_ok = True
except ImportError:
    print("This script must be run under WeeChat.")
    print("Get WeeChat now at: http://www.weechat.org/")
    import_ok = False

SCRIPT_NAME    = "completion"
SCRIPT_AUTHOR  = "Elián Hanisch <lambdae2@gmail.com>"
SCRIPT_VERSION = "0.3"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC    = "Word completions for WeeChat"
SCRIPT_COMMAND = "completion"

### Config ###
settings = {
'replace_values':''
}

### Messages ###
def debug(s, prefix='', buffer=None):
    """Debug msg"""
    #if not weechat.config_get_plugin('debug'): return
    if buffer is None:
        buffer_name = 'DEBUG_' + SCRIPT_NAME
        buffer = weechat.buffer_search('python', buffer_name)
        if not buffer:
            buffer = weechat.buffer_new(buffer_name, '', '', '', '')
            weechat.buffer_set(buffer, 'nicklist', '0')
            weechat.buffer_set(buffer, 'time_for_each_line', '0')
            weechat.buffer_set(buffer, 'localvar_set_no_log', '1')
    weechat.prnt(buffer, '%s\t%s' %(prefix, s))

def error(s, prefix=None, buffer='', trace=''):
    """Error msg"""
    prefix = prefix or script_nick
    weechat.prnt(buffer, '%s%s %s' %(weechat.prefix('error'), prefix, s))
    if weechat.config_get_plugin('debug'):
        if not trace:
            import traceback
            if traceback.sys.exc_type:
                trace = traceback.format_exc()
        not trace or weechat.prnt('', trace)

def say(s, prefix=None, buffer=''):
    """normal msg"""
    prefix = prefix or script_nick
    weechat.prnt(buffer, '%s\t%s' %(prefix, s))

print_replace = lambda k,v : say('%s %s=>%s %s' %(k, color_delimiter, color_reset, v))

### Config functions ###
def get_config_dict(config):
    value = weechat.config_get_plugin(config)
    if not value:
        return {}
    values = [s.split('=>') for s in value.split(';;')]
    #debug(values)
    return dict(values)

def load_replace_table():
    global replace_table
    replace_table = dict(get_config_dict('replace_values'))

def save_replace_table():
    global replace_table
    weechat.config_set_plugin('replace_values',
            ';;'.join(['%s=>%s' %(k, v) for k, v in replace_table.items()]))

### Commands ###
def cmd_completion(data, buffer, args):
    global replace_table
    if not args:
        if replace_table:
            for k, v in replace_table.items():
                print_replace(k, v)
        else:
            say('No completions.')
        return WEECHAT_RC_OK
    cmd, space, args = args.partition(' ')
    if cmd == 'add':
        word, space, text = args.partition(' ')
        k, v = word.strip(), text.strip()
        replace_table[k] = v
        save_replace_table()
        say('added: %s %s=>%s %s' %(k, color_delimiter, color_reset, v))
    elif cmd == 'del':
        k = args.strip()
        try:
            del replace_table[k]
            save_replace_table()
            say("completion for '%s' deleted." %k)
            save_replace_table()
        except KeyError:
            error("completion for '%s' not found." %k)
    return WEECHAT_RC_OK

### Completion ###
def completion_replacer(data, completion_item, buffer, completion):
    global replace_table
    pos = weechat.buffer_get_integer(buffer, 'input_pos')
    input = weechat.buffer_get_string(buffer, 'input')
    #debug('%r %s %s' %(input, len(input), pos))
    if pos > 0 and (pos == len(input) or input[pos] == ' '):
        n = input.rfind(' ', 0, pos)
        word = input[n+1:pos]
        #debug(repr(word))
        if word in replace_table:
            replace = replace_table[word]
            if pos >= len(input.strip()):
                # cursor is in the end of line, append a space
                replace += ' '
            n = len(word)
            input = '%s%s%s' %(input[:pos-n], replace, input[pos:])
            weechat.buffer_set(buffer, 'input', input)
            weechat.buffer_set(buffer, 'input_pos', str(pos - n + len(replace)))
    return WEECHAT_RC_OK

def completion_keys(data, completion_item, buffer, completion):
    global replace_table
    for k in replace_table:
        weechat.hook_completion_list_add(completion, k, 0, weechat.WEECHAT_LIST_POS_SORT)
    return WEECHAT_RC_OK

### Main ###
if __name__ == '__main__' and import_ok and \
        weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, \
        SCRIPT_DESC, '', ''):

    # colors
    color_delimiter   = weechat.color('chat_delimiters')
    color_script_nick = weechat.color('chat_nick')
    color_reset   = weechat.color('reset')

    # pretty [SCRIPT_NAME]
    script_nick = '%s[%s%s%s]%s' %(color_delimiter, color_script_nick, 'cmpl', color_delimiter,
            color_reset)

    version = weechat.info_get('version', '')
    if version == '0.3.0':
        error('WeeChat 0.3.1 or newer is required for this script.')
    else:
        # settings
        for opt, val in settings.items():
            if not weechat.config_is_set_plugin(opt):
                weechat.config_set_plugin(opt, val)

        load_replace_table()

        completion_template = 'completion_script'
        weechat.hook_completion(completion_template,
                "Replaces last word in input by its configured value.", 'completion_replacer', '')
        weechat.hook_completion('completion_keys', "Words in completion list.", 'completion_keys', '')

        weechat.hook_command(SCRIPT_COMMAND, SCRIPT_DESC , "[add <word> <text>|del <word>]",
"""
add: adds a new completion, <word> => <text>.
del: deletes a completion.
Without arguments it displays current completions.

<word> will be replaced by <text> when pressing tab in input line,
where <word> is any word currently behind the cursor.

Setup:
For this script to work, you must add the template
%%(%(completion)s) to the default completion template, use:
/set weechat.completion.default_template "%%(nicks)|%%(irc_channels)|%%(%(completion)s)"

Examples:
/%(command)s add wee WeeChat (typing wee<tab> will replace 'wee' by 'WeeChat')
/%(command)s add weeurl http://www.weechat.org/
/%(command)s add test This is a test!
""" %dict(completion=completion_template, command=SCRIPT_COMMAND),
            'add|del %(completion_keys)', 'cmd_completion', '')

# vim:set shiftwidth=4 tabstop=4 softtabstop=4 expandtab textwidth=100:
